#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import tarfile

import pysam

import paleomix.yaml

from paleomix.common.formats.fasta import FASTA
from paleomix.tools.zonkey.common import \
    get_sample_names, \
    contig_name_to_plink_name

from paleomix.ui import \
    print_warn, \
    print_info, \
    print_err


_SETTINGS_KEYS = ('Format', 'Revision', 'Plink', 'NChroms', 'MitoPadding',
                  'SNPDistance')


class BAMInfo(object):
    def __init__(self):
        self.nuclear = False
        self.mt_contig = None
        self.mt_length = None
        self.mt_padding = None

    @property
    def is_nuclear(self):
        return self.nuclear

    @property
    def is_mitochondrial(self):
        return bool(self.mt_contig)

    def __repr__(self):
        tmpl = "BAMInfo(nuclear=%r, mt_contig=%r, mt_length=%r, mt_padding=%r)"

        return tmpl % (self.nuclear, self.mt_contig,
                       self.mt_length, self.mt_padding)


# Format number for database file; is incremented when the format is changed.
# The 'revision' field specifies updates to the table that do not change the
# format of the database (see below).
_SUPPORTED_DB_FORMAT_MAJOR = 1
_SUPPORTED_DB_FORMAT_MINOR = 20160112

# Required columns in the 'contigs.txt' table; additional columns are ignored
_CONTIGS_TABLE_COLUMNS = frozenset(('ID', 'Size', 'Checksum'))
# Required columns in the 'samples.txt' table; additional columns are ignored
_SAMPELS_TABLE_COLUMNS = frozenset(('ID', 'Group(2)', 'Group(3)', 'Species',
                                    'Sex', 'SampleID', 'Publication'))


class ZonkeyDBError(RuntimeError):
    pass


class ZonkeyDB(object):
    def __init__(self, filename):
        self.filename = filename

        with tarfile.open(filename) as tar_handle:
            self.settings = self._read_settings(tar_handle, "settings.yaml")
            self.contigs = self._read_contigs_table(tar_handle, "contigs.txt")
            self.samples = self._read_samples_table(tar_handle, "samples.txt")
            self.sample_order = self._read_sample_order(tar_handle,
                                                        "genotypes.txt")
            self.mitochondria = self._read_mitochondria(tar_handle,
                                                        "mitochondria.fasta")
            self.simulations = self._read_simulations(tar_handle,
                                                      "simulations.txt")

        self._cross_validate()

    def validate_bam(self, filename):
        """Validates a sample BAM file, checking that it is either a valid
        mitochondrial BAM (aligned against one of the referenc mt sequences),
        or that it is a valid nuclear BAM (aligned against the reference).

        Returns one of INVALID_BAMFILE, NUC_BAMFILE, and MITO_BAMFILE.
        """
        print_info("  - Validating BAM file %r ... " % (filename,))

        try:
            handle = pysam.Samfile(filename)
        except (ValueError, IOError), error:
            print_err("Error reading BAM: %s" % (error,))
            return

        return self.validate_bam_handle(handle)

    def validate_bam_handle(self, handle):
        samples = get_sample_names(handle)
        if len(samples) > 1:
            print_warn("\nWARNING:")
            print_warn("BAM read-groups specify more than one sample, "
                       "but this tool treats BAMs as a single sample:")

            for sample in enumerate(samples, start=1):
                print_warn("    %i: %r" % sample)
            print_warn("")

        info = BAMInfo()
        if not _validate_mito_bam(self, handle, info):
            return

        if not _validate_nuclear_bam(self, handle, info):
            return

        return info

    def _cross_validate(self):
        """Cross validates tables to ensure consistency."""
        genotypes = set(self.sample_order)
        samples = set(self.samples)
        differences = (genotypes | samples) - (genotypes & samples)
        if differences:
            raise ZonkeyDBError("Mismatch between samples in sample-list and "
                                "genotypes table; some samples not found in "
                                "both tables: %s"
                                % (",".join(differences),))

        for name, record in self.mitochondria.iteritems():
            if name not in self.samples:
                # Ignore extra reference sequences
                meta = (record.meta or "").upper()
                if "EXCLUDE" not in map(str.strip, meta.split(";")):
                    raise ZonkeyDBError("Unexpected mitochondrial sequence: %r"
                                        % (name,))

    @classmethod
    def _read_contigs_table(cls, tar_handle, filename):
        cls._check_required_file(tar_handle, filename)

        table = cls._read_table(tar_handle, filename, _CONTIGS_TABLE_COLUMNS)
        for key, row in table.iteritems():
            try:
                row["Size"] = int(row["Size"])
            except ValueError, error:
                raise ZonkeyDBError("Invalid size specified for sample %r in "
                                    "%r: %r" % (key, filename, error))

            if row["Size"] <= 0:
                raise ZonkeyDBError("Contig size must be >= 0 for %r in %r, "
                                    "not %r" % (key, filename, row["Size"]))
        return table

    @classmethod
    def _read_samples_table(cls, tar_handle, filename):
        cls._check_required_file(tar_handle, filename)

        samples = cls._read_table(tar_handle, "samples.txt")
        if not samples:
            raise ZonkeyDBError("ERROR: No samples found in genotypes table!")

        for row in samples.itervalues():
            if row["Sex"].upper() not in ("MALE", "FEMALE", "NA"):
                raise ZonkeyDBError("ERROR: Unexpected sample sex (%r); "
                                    "expected 'MALE', 'FEMALE', or 'NA'"
                                    % (row["Sex"],))

        for k_groups in (2, 3):
            key = "Group(%i)" % (k_groups,)
            groups = frozenset(row[key] for row in samples.itervalues())

            if len(groups - set('-')) not in (0, k_groups):
                raise ZonkeyDBError("The %r column in the samples table must "
                                    "either contain %i ancestral groups, or "
                                    "none" % (key, k_groups))

        return samples

    @classmethod
    def _read_sample_order(cls, tar_handle, filename):
        cls._check_required_file(tar_handle, filename)

        handle = tar_handle.extractfile(filename)
        header = handle.readline().rstrip('\r\n').split('\t')
        sample_order = tuple(header[-1].split(";"))

        if len(sample_order) != len(set(sample_order)):
            raise ZonkeyDBError("Duplicate sample names in %r" % (filename,))

        return sample_order

    def _read_mitochondria(self, tar_handle, filename):
        try:
            tar_handle.getmember(filename)
        except KeyError:
            # Missing MT file is allowed
            return None

        handle = tar_handle.extractfile(filename)

        results = {}
        for record in FASTA.from_lines(handle):
            record = FASTA(name=record.name,
                           meta=record.meta,
                           sequence=record.sequence.upper())

            unexpected = set(record.sequence) - set("ACGTN-")
            if unexpected:
                unexpected = ", ".join(map(repr, sorted(unexpected)))
                raise ZonkeyDBError("Unexpected nucleotide in %s; only A, C, "
                                    "G, T, N, and - are allowed, not %s"
                                    % (unexpected, filename))
            elif record.name in results:
                raise ZonkeyDBError("Duplicate sequence name in %s: %r"
                                    % (filename, record.name))

            results[record.name] = record

        lengths = frozenset(len(record.sequence)
                            for record in results.itervalues())

        if not lengths:
            raise ZonkeyDBError("No mitochondrial sequences found in %r"
                                % (filename,))
        elif len(lengths) > 2:
            lengths = tuple(sorted(lengths))
            lengths_s = "%s, and %s" % (", ".join(map(str, lengths[:-1])),
                                        lengths[-1])

            raise ZonkeyDBError("At most two different sequence lengths "
                                "expected for mitochondrial sequences, but "
                                "found %i different lengths in %r: %s"
                                % (len(lengths), filename, lengths_s))
        elif len(lengths) != 1:
            # Unpadded sequences are allowed
            delta_len = max(lengths) - min(lengths)
            mito_padding = self.settings["MitoPadding"]

            if (delta_len != mito_padding):
                raise ZonkeyDBError("Length difference between mitochondrial "
                                    "sequences in %r does not match the "
                                    "padding; expected a difference of %i bp, "
                                    "but found a %i bp difference."
                                    % (filename, mito_padding, delta_len))

        return results

    @classmethod
    def _read_settings(cls, tar_handle, filename):
        cls._check_required_file(tar_handle, filename)

        handle = tar_handle.extractfile(filename)

        try:
            result = paleomix.yaml.safe_load(handle.read())
        except paleomix.yaml.YAMLError, error:
            raise ZonkeyDBError("Error reading settings file %r; %s"
                                % (filename, error))

        for key in _SETTINGS_KEYS:
            if key != "Plink":
                if not isinstance(result[key], int) or result[key] < 0:
                    raise ZonkeyDBError("Value for %r in %s must be an non-"
                                        "negative integer, not %r"
                                        % (key, filename, result[key]))
            elif not isinstance(result[key], str):
                    raise ZonkeyDBError("Value for %r in %s must be a string, "
                                        "not %r"
                                        % (key, filename, result[key]))

        if result["Format"] > _SUPPORTED_DB_FORMAT_MAJOR:
            raise ZonkeyDBError("Database version is too old; this version of "
                                "PALEOMIX supports the Zonkey DB v%i, but the "
                                "database is v%i; download an updated "
                                "database to continue."
                                % (_SUPPORTED_DB_FORMAT_MAJOR,
                                   result["Format"]))
        elif result["Format"] < _SUPPORTED_DB_FORMAT_MAJOR:
            raise ZonkeyDBError("Database version is too new; this version of "
                                "PALEOMIX supports the Zonkey DB v%i, but the "
                                "database is v%i; upgrade PALEOMIX to "
                                "continue."
                                % (_SUPPORTED_DB_FORMAT_MAJOR,
                                   result["Format"]))
        elif result["Revision"] < _SUPPORTED_DB_FORMAT_MINOR:
            raise ZonkeyDBError("Database version is too old; this version of "
                                "PALEOMIX supports the Zonkey DB v%i, rev. %i "
                                "or newer, but the database is v%i rev. %i; "
                                "please download an updated database to "
                                "continue."
                                % (_SUPPORTED_DB_FORMAT_MAJOR,
                                   _SUPPORTED_DB_FORMAT_MINOR,
                                   result["Format"],
                                   result["Revision"]))

        return result

    def _read_simulations(self, tar_handle, filename):
        self._check_required_file(tar_handle, filename)

        handle = tar_handle.extractfile(filename)
        header = handle.readline().rstrip().split('\t')

        required_keys = set(('NReads', 'K', 'Sample1', 'Sample2', 'HasTS',
                             'Percentile', 'Value'))
        missing_keys = required_keys - set(header)
        if missing_keys:
            missing_keys = ', '.join(map(repr, missing_keys))
            raise ZonkeyDBError('Simulations table %r does not contain all '
                                'required columns; columns %r are missing!'
                                % (filename, missing_keys))

        result = []
        for linenum, line in enumerate(handle, start=2):
            fields = line.strip().split('\t')
            if len(fields) != len(header):
                raise ZonkeyDBError("Line %i in simulations table %r, does "
                                    "not contain the expected number of "
                                    "columns; expected %i, but found %i!"
                                    % (linenum, filename,
                                       len(header), len(fields)))

            row = dict(zip(header, fields))

            if row['HasTS'] not in ('TRUE', 'FALSE'):
                pass

            row['HasTS'] = (row['HasTS'] == 'TRUE')

            for key in ('NReads', 'K'):
                try:
                    row[key] = int(row[key])
                except ValueError:
                    raise ZonkeyDBError('Malformed value for column %r at '
                                        'line %i in simulations table %r; '
                                        'expected int, found %r'
                                        % (key, linenum, filename, row[key]))

            for key in ('Percentile', 'Value'):
                try:
                    row[key] = float(row[key])
                except ValueError:
                    raise ZonkeyDBError('Malformed value for column %r at '
                                        'line %i in simulations table %r; '
                                        'expected float, found %r'
                                        % (key, linenum, filename, row[key]))

            for key in ('Sample1', 'Sample2'):
                group_key = 'Group(%i)' % (row['K'],)
                groups = frozenset(row[group_key]
                                   for row in self.samples.itervalues())

                if row[key] not in groups and row[key] != '-':
                    raise ZonkeyDBError('Invalid group in column %r in '
                                        'simulations table %r: %r'
                                        % (key, filename, row[key]))

            result.append(row)

        return result

    @classmethod
    def _check_required_file(cls, tar_handle, filename):
        try:
            obj = tar_handle.getmember(filename)
        except KeyError:
            raise ZonkeyDBError("Database does not contain required file %r; "
                                "please ensure that this is a valid Zonkey "
                                "database file!" % (filename,))

        if not obj.isfile():
            raise ZonkeyDBError("Object %r in Zonkey database is not a "
                                "file; please ensure that this is a valid "
                                "Zonkey database file!"
                                % (filename,))

    @classmethod
    def _read_table(cls, tar_handle, filename, requied_columns=()):
        requied_columns = frozenset(requied_columns) | frozenset(("ID",))
        handle = tar_handle.extractfile(filename)
        result = {}

        try:
            header = handle.readline().rstrip('\r\n').split('\t')
            if len(header) != len(set(header)):
                raise ZonkeyDBError("Table %r does contains duplicate columns!"
                                    % (filename,))

            if requied_columns - set(header):
                raise ZonkeyDBError("Required columns are missign in table "
                                    "%r: %s" % (filename, ", ".join()))

            for linenum, line in enumerate(handle):
                fields = line.rstrip('\r\n').split('\t')

                if len(fields) != len(header):
                    raise ZonkeyDBError("Error reading %r at line %i; "
                                        "expected  %i columns, found %i "
                                        "columns!"
                                        % (filename, linenum,
                                           len(header), len(fields)))

                row = dict(zip(header, fields))
                if row["ID"] in result:
                    raise ZonkeyDBError("Duplicate IDs in %r: %s"
                                        % (filename, row["ID"]))

                result[row["ID"]] = row
        finally:
            handle.close()

        return result


def _validate_mito_bam(data, handle, info):
    references = handle.references
    min_length = min((len(record.sequence))
                     for record in data.mitochondria.itervalues())

    for bam_contig, bam_length in zip(references, handle.lengths):
        if bam_contig not in data.mitochondria:
            continue

        db_sequence = data.mitochondria[bam_contig].sequence
        db_length = len(db_sequence) - db_sequence.count("-")

        if bam_length != db_length:
            print_err("ERROR: Length of mitochondrial contig %r (%i bp) "
                      "does not match the length of the corresponding "
                      "sequence in the database (%i bp)"
                      % (bam_contig, bam_length, db_length))
            return False

        for line in pysam.idxstats(handle.filename):
            name, _, hits, _ = line.strip().split('\t')
            if (name == bam_contig) and not int(hits):
                print_err("ERROR: Mitochondrial BAM (%r) does not contain "
                          "any reads aligned to contig %r; inferring an "
                          "phylogeny is not possible."
                          % (handle.filename, name))
                return False

        info.mt_contig = bam_contig
        info.mt_length = bam_length
        info.mt_padding = len(db_sequence) - min_length

        return True
    return True


def _validate_nuclear_bam(data, handle, info):
    # Check that chromosomes are of expected size; unused chroms are ignored.
    bam_contigs = dict(zip(map(contig_name_to_plink_name, handle.references),
                           handle.lengths))
    ref_contigs = data.contigs

    contigs_found = {}
    for name, stats in sorted(ref_contigs.iteritems()):
        if name not in bam_contigs:
            contigs_found[name] = False
        elif bam_contigs[name] != stats["Size"]:
            print_err("\nERROR: Chrom %r in the BAM does not match the "
                      "length specified in data file:\n"
                      "    - Expected: %i\n"
                      "    - Found: %i"
                      % (name, bam_contigs[name], stats["Size"]))

            return False
        else:
            contigs_found[name] = True

    if any(contigs_found.itervalues()):
        if not all(contigs_found.itervalues()):
            print_err("\nERROR: Not all nuclear chromosomes found in BAM:")
            for (name, stats) in sorted(ref_contigs.iteritems()):
                is_found = "Found" if contigs_found[name] else "Not found!"
                print_err("  - %s: %s" % (name, is_found))

            return False
        else:
            info.nuclear = True

    return True



#!/usr/bin/env python

"""
Stripped down version of 'FilterUniqueBAM' by
:Author: Martin Kircher
:Contact: Martin.Kircher@eva.mpg.de
:Date: *08.10.2011
:Type: tool
:Input: BAM
:Output: BAM

Mark/Filter PCR duplicates for merged PE reads Reads BAM
from STDIN and writes BAM to STDOUT. All non-collapsed reads
as well as secondary/chinermic alignments, reads that have
failed QC and unmmaped reads, are written to STDOUT as is.

The input is assumed to be sorted by coordinates, and this
order is preservered, though individual reads at the same
position may be re-arranged).
"""

import sys
import pysam
import random

from argparse import ArgumentParser


def calc_consensus(reads, rng=random.random):
    count = len(reads)
    outread = None
    maxsumqual = -1
    for read in reads:
        # Handle reads without qualities, but favor reads with qualities
        qual = read.qual
        if qual is None:
            # Generate value in (-1; 0]
            nsum = -rng()
        else:
            nsum = sum(map(ord, qual))

        if nsum > maxsumqual:
            outread = read
            maxsumqual = nsum

        # LOOK FOR PREVIOUS PCR DUPLICATE COUNTS
        for key, value in read.tags:
            if key == "XP":
                count += value

    if not outread.tags:
        outread.tags = [("XP", count)]
    else:
        outread.tags = outread.tags + [("XP", count)]

    return outread


def get_consensus_se(reads):
    # DETERMINE MOST FREQUENT CIGAR LINE
    by_cigar = {}
    cigar_count = {}
    for read in reads:
        tcigar = tuple(read.cigar)
        if tcigar in by_cigar:
            cigar_count[tcigar] += 1
            by_cigar[tcigar].append(read)
        else:
            cigar_count[tcigar] = 1
            by_cigar[tcigar] = [read]

    to_sort = [(y, -len(str(x)), x) for (x, y) in cigar_count.iteritems()]
    to_sort.sort()
    selcigar = to_sort[-1][-1]
    reads = by_cigar[selcigar]

    return calc_consensus(reads)


def write_consensus_se(outfile, reads, remove_duplicates):
    consensus = get_consensus_se(reads)
    for read in reads:
        read.is_duplicate = (read is not consensus)
        if not (read.is_duplicate and remove_duplicates):
            outfile.write(read)


def _flush_buffer(outfile, curvariants, remove_duplicates):
    for value in curvariants.itervalues():
        write_consensus_se(outfile, value[0], remove_duplicates)
    curvariants.clear()


_FILTERED_FLAGS = 0x1     # PE reads
_FILTERED_FLAGS |= 0x4    # Unmapped
_FILTERED_FLAGS |= 0x100  # Secondary alignment
_FILTERED_FLAGS |= 0x200  # Failed QC
_FILTERED_FLAGS |= 0x800  # Chimeric alignment


def parse_args(argv):
    usage = """paleomix rmdup_collapsed [options] < sorted.bam > out.bam

The rmdup_collapsed filters a BAM file for PCR duplicates unpaired reads under
the assumption that any unpaired read have been generated by the merging of
overlapping paired-end reads, and thereby represent the complete template
sequence. PCR duplicates are therefore detected based on both the 5' and 3'
alignment coordinate.

Paired reads (0x1), unmapped reads (0x4), secondary alignments (0x100),
reads that failed QC (0x200), and chimeric alignments (0x800), as identified
using the BAM record flags, are not filtered, but simply written to the output.

By default, filtered reads are flagged using the "duplicate" flag (0x400), and
written to the output. Use the --remove-duplicates command-line option to
instead remove these records from the output.
"""
    parser = ArgumentParser(usage=usage)
    parser.add_argument("input", default="-", nargs="?",
                        help="BAM file; if not set, input is read from STDIN.")
    parser.add_argument("--remove-duplicates",
                        help="Remove duplicates from output; by default "
                             "duplicates are only flagged (flag = 0x400).",
                        default=False, action="store_true")
    parser.add_argument("--seed", default=None, type=int,
                        help="Seed used for randomly selecting representative "
                             "reads when no reads have quality scores assigned"
                             "[default: initialized using system time].")

    return parser.parse_args(argv)


def main(argv):
    args = parse_args(argv)

    # Initialize seed used when selecting among reads without quality scores
    random.seed(args.seed)

    if args.input == "-" and sys.stdin.isatty():
        sys.stderr.write("STDIN is a terminal, terminating!\n")
        return 1
    elif sys.stdout.isatty():
        sys.stderr.write("STDOUT is a terminal, terminating!\n")
        return 1

    with pysam.Samfile(args.input, "rb") as infile:
        with pysam.Samfile("-", "wb", template=infile) as outfile:
            curpos = None
            curvariants = {}
            for (read_num, read) in enumerate(infile):
                if curpos and ((read.tid, read.pos) != curpos):
                    # Sort order is defined as ascending 'tid's and positions
                    if curpos > (read.tid, read.pos) and not read.is_unmapped:
                        sys.stderr.write("ERROR: Input file does not appear "
                                         "to be sorted by coordinates at "
                                         "record %i, aborting ...\n"
                                         % (read_num,))
                        return 1

                    _flush_buffer(outfile, curvariants,
                                  args.remove_duplicates)
                    curpos = None

                if read.flag & _FILTERED_FLAGS:
                    outfile.write(read)
                    continue

                curpos = (read.tid, read.pos)
                nkey = (read.is_reverse, read.pos, read.alen)
                if nkey in curvariants:
                    curvariants[nkey][0].append(read)
                    curvariants[nkey][1] += 1
                else:
                    curvariants[nkey] = [[read], 1]

            _flush_buffer(outfile, curvariants, args.remove_duplicates)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import sys
import datetime
import collections

from paleomix.common.utilities import \
    get_in, \
    set_in
from paleomix.common.text import \
    padded_table, \
    parse_padded_table

from paleomix.tools.bam_stats.common import \
    BAMStatsError


##############################################################################
##############################################################################
##

READGROUP_TEMPLATE = {"SE": 0, "PE_1": 0, "PE_2": 0, "Collapsed": 0,
                      "Hits": 0, "M": 0, "I": 0, "D": 0, "Size": 0}


# Header prepended to output tables
TABLE_HEADER = """# Timestamp: %s
#
# Columns:
#   Contig:   Contig, chromosome, or feature for which a depth histogram was
#             created. Unnamed features are named after the chromosome or
#             contig on which they are located, with a star appended. For
#             example "chr1*".
#   Size:     The total size of the region. Multiple features with the same
#             name are combined into one row, with the size representing to
#             total of these. Note that overlapping bases are counted 2 (or
#             more) times.
#  Hits:      Sum of SE, PE_1, and PE_2 hits. Note that supplementary
#             alignments, duplicates, reads that failed QC, secondary
#             alignments, and unmapped reads are ignored.
#  SE, PE_*:  Number of Single Ended, and Pair Ended (mate 1 and 2) hits
#             overlapping the current contig or intervals. Note that a hit may
#             be counted multiple times if it overlaps multiple intervals
#  Collapsed: Number of hits for PE pair collapsed into a single read
#  M, I, D:   Number of aligned (M), inserted (I) and deleted (D) bases
#             relative to references
#  Coverage:  Average number of bases covering each position in the
#             contig(s)/intervals(s).
"""


def calculate_totals(table):
    lengths = {}
    for samples in table.itervalues():
        for libraries in samples.values():
            for contigs in libraries.values():
                for (name, contig) in contigs.iteritems():
                    size = lengths.get(name)
                    if (size is not None) and (size != contig["Size"]):
                        raise BAMStatsError(name)
                    lengths[name] = contig["Size"]

    for (name, samples) in sorted(table.items()):
        for (sample, libraries) in sorted(samples.items()):
            for (library, contigs) in sorted(libraries.items()):
                totals = _calculate_totals_in(contigs, lengths)
                set_in(table, (name, sample, library), totals)

            totals = _calculate_totals_in(libraries, lengths)
            set_in(table, (name, sample, "*"), totals)

        set_in(table, (name, "*", "*"), _calculate_totals_in(table, lengths))
    return table


def build_rows(table):
    rows = [("Name", "Sample", "Library", "Contig", "Size", "Hits", "SE",
             "PE_1", "PE_2", "Collapsed", "M", "I", "D", "Coverage")]

    for (name, samples) in sorted(table.items()):
        for (sample, libraries) in sorted(samples.items()):
            for (library, contigs) in sorted(libraries.items()):
                for (contig, subtable) in sorted(contigs.items()):
                    row = [name,
                           sample,
                           library,
                           contig,
                           subtable["Size"],
                           subtable["SE"] + subtable["PE_1"]
                                          + subtable["PE_2"]
                                          + subtable["Collapsed"],
                           subtable["SE"],
                           subtable["PE_1"],
                           subtable["PE_2"],
                           subtable["Collapsed"],
                           subtable["M"],
                           subtable["I"],
                           subtable["D"],
                           float(subtable["M"]) / subtable["Size"]]
                    rows.append(row)
                rows.append("#")
            rows.append("#")

    while rows[-1] == "#":
        rows.pop()
    return rows


def read_table(table, filename):
    with open(filename) as table_file:
        for record in parse_padded_table(table_file):
            key = (record["Name"], record["Sample"],
                   record["Library"], record["Contig"])
            if "*" in key:
                continue

            subtable = get_in(table, key)
            if subtable is None:
                subtable = dict(READGROUP_TEMPLATE)
                subtable["Size"] = int(record["Size"])
                set_in(table, key, subtable)

            assert int(subtable["Size"]) == int(record["Size"])
            for key in READGROUP_TEMPLATE:
                if key != "Size":
                    subtable[key] += int(record.get(key, 0))


def write_table(table, filename):
    table = calculate_totals(table)
    rows = build_rows(table)

    if filename == "-":
        output_handle = sys.stdout
    else:
        output_handle = open(filename, "w")

    try:
        output_handle.write(TABLE_HEADER % datetime.datetime.now().isoformat())
        for line in padded_table(rows):
            output_handle.write(line)
            output_handle.write("\n")
    finally:
        if output_handle is not sys.stdout:
            output_handle.close()


def _calculate_totals_in(tables, lengths):
    def _defaults():
        return dict(READGROUP_TEMPLATE)

    totals = collections.defaultdict(_defaults)
    total_size = sum(lengths.itervalues())

    subtables = tables.items()
    while subtables:
        subtable_key, subtable = subtables.pop()
        if subtable_key == "*":
            totals[subtable_key]["Size"] = total_size
        elif "SE" in subtable:
            for key in READGROUP_TEMPLATE:
                if key != "Size":
                    totals[subtable_key][key] += subtable[key]
                    totals["*"][key] += subtable[key]
                else:
                    totals[subtable_key][key] = lengths[subtable_key]
                    totals["*"][key] = total_size
        else:
            subtables.extend(subtable.items())

    return dict(totals)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Implements nodes for calling PALEOMIX commands.

Each node is equivalent to a particular command:
    $ paleomix [...]
"""
import os
import gzip

from paleomix.node import \
    CommandNode, \
    Node
from paleomix.atomiccmd.command import \
    AtomicCmd
from paleomix.atomiccmd.sets import \
    ParallelCmds
from paleomix.common.fileutils import \
    describe_files
from paleomix.nodes.picard import \
    MultiBAMInput, \
    MultiBAMInputNode
from paleomix.atomiccmd.builder import \
    AtomicCmdBuilder, \
    create_customizable_cli_parameters, \
    use_customizable_cli_parameters
from paleomix.common.fileutils import \
    reroot_path, \
    move_file

from paleomix.nodes.samtools import \
    SAMTOOLS_VERSION, \
    SAMTOOLS_VERSION_0119, \
    BCFTOOLS_VERSION_0119

import paleomix.tools.bam_stats.coverage as coverage
import paleomix.tools.factory as factory


class DuplicateHistogramNode(MultiBAMInputNode):
    """Node for calling the 'paleomix duphist' command.

    Takes 1 or more BAMs as imput, requiring a config object in order to run
    MergeSamFiles.jar to merge these files. The output is a histogram of PCR
    duplicate counts, usable as input for the 'preseq' tool.
    """

    def __init__(self, config, input_files, output_file, dependencies=()):
        bam_input = MultiBAMInput(config, input_files)
        duphist_command = factory.new("duphist")
        duphist_command.add_value('%(TEMP_IN_BAM)s')
        duphist_command.set_kwargs(OUT_STDOUT=output_file)
        bam_input.setup(duphist_command)
        duphist_command = duphist_command.finalize()

        commands = ParallelCmds(bam_input.commands + [duphist_command])

        description = "<DuplicateHistogram: %s -> %r>" \
            % (describe_files(input_files), output_file)
        MultiBAMInputNode.__init__(self,
                                   bam_input=bam_input,
                                   command=commands,
                                   description=description,
                                   dependencies=dependencies)


class CoverageNode(CommandNode):
    def __init__(self, config, target_name, input_file, output_file,
                 regions_file=None, dependencies=()):
        builder = factory.new("coverage")
        builder.add_value("%(IN_BAM)s")
        builder.add_value("%(OUT_FILE)s")
        builder.set_option("--target-name", target_name)
        builder.set_kwargs(IN_BAM=input_file,
                           OUT_FILE=output_file)

        if regions_file:
            builder.set_option('--regions-file', '%(IN_REGIONS)s')
            builder.set_kwargs(IN_REGIONS=regions_file)

        description = "<Coverage: %s -> '%s'>" % (input_file, output_file)
        CommandNode.__init__(self,
                             command=builder.finalize(),
                             description=description,
                             dependencies=dependencies)


class MergeCoverageNode(Node):
    def __init__(self, input_files, output_file, dependencies=()):
        self._output_file = output_file

        Node.__init__(self,
                      description="<MergeCoverage: '%s' -> '%s'>"
                      % (describe_files(input_files), self._output_file),
                      input_files=input_files,
                      output_files=self._output_file,
                      dependencies=dependencies)

    def _run(self, _config, temp):
        table = {}
        for filename in self.input_files:
            coverage.read_table(table, filename)

        coverage.write_table(table, reroot_path(temp, self._output_file))
        move_file(reroot_path(temp, self._output_file), self._output_file)


class DepthHistogramNode(MultiBAMInputNode):
    def __init__(self, config, target_name, input_files, output_file,
                 regions_file=None, dependencies=()):
        bam_input = MultiBAMInput(config, input_files)
        if len(bam_input.files) > 1 and regions_file:
            raise ValueError("DepthHistogram for regions require single, "
                             "indexed input BAM file.")

        builder = factory.new("depths")
        builder.add_value("%(TEMP_IN_BAM)s")
        builder.add_value("%(OUT_FILE)s")
        builder.set_option("--target-name", target_name)
        builder.set_kwargs(OUT_FILE=output_file)
        bam_input.setup(builder)

        if regions_file:
            builder.set_option('--regions-file', '%(IN_REGIONS)s')
            builder.set_kwargs(IN_REGIONS=regions_file)

        command = ParallelCmds(bam_input.commands + [builder.finalize()])
        description = "<DepthHistogram: %s -> '%s'>" \
            % (describe_files(bam_input.files), output_file)
        MultiBAMInputNode.__init__(self,
                                   bam_input=bam_input,
                                   command=command,
                                   description=description,
                                   dependencies=dependencies)


class FilterCollapsedBAMNode(MultiBAMInputNode):
    def __init__(self, config, input_bams, output_bam, keep_dupes=True,
                 dependencies=()):
        bam_input = MultiBAMInput(config, input_bams, indexed=False)

        builder = factory.new("rmdup_collapsed")
        builder.add_value("%(TEMP_IN_BAM)s")
        builder.set_kwargs(OUT_STDOUT=output_bam)
        bam_input.setup(builder)

        if not keep_dupes:
            builder.set_option("--remove-duplicates")

        filteruniq = builder.finalize()
        command = ParallelCmds(bam_input.commands + [filteruniq])
        description = "<FilterCollapsedBAM: %s>" \
            % (describe_files(bam_input.files),)
        MultiBAMInputNode.__init__(self,
                                   bam_input=bam_input,
                                   command=command,
                                   description=description,
                                   dependencies=dependencies)


class VCFPileupNode(CommandNode):
    """Collects heterozygous SNPs from a VCF file, and generates a bgzipped
    pileup for those sites containing the SNPs.

    The resulting pileup is read by 'paleomix vcf_filter'; this allows
    filtering based on the frequency of the minority SNP, since this is not
    reported in the VCF.
    """

    @create_customizable_cli_parameters
    def customize(cls, reference, infile_bam, infile_vcf, outfile,
                  dependencies=()):
        params = factory.new("genotype")
        params.add_value("%(IN_BAMFILE)s")
        params.add_value("%(OUT_PILEUP)s")
        params.set_option("--bedfile", "%(TEMP_IN_INTERVALS)s")
        params.set_option("--pileup-only")
        # Ignore read-groups for pileup
        params.add_option("--mpileup-argument", "-R", sep="=")
        # Reference sequence (FASTA)
        params.add_option("--mpileup-argument",
                          "-f=%s" % (reference,), sep="=")

        params.set_kwargs(IN_BAMFILE=infile_bam,
                          TEMP_IN_INTERVALS="heterozygous_snps.bed",
                          # Automatically remove this file
                          TEMP_OUT_INTERVALS="heterozygous_snps.bed",
                          OUT_PILEUP=outfile,
                          CHECK_SAMTOOLS=SAMTOOLS_VERSION)

        return {"command": params}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        self._in_vcf = parameters.infile_vcf
        command = parameters.command.finalize()
        description = "<VCFPileup: '%s' -> '%s'>" \
            % (parameters.infile_vcf,
               parameters.outfile)

        CommandNode.__init__(self,
                             description=description,
                             command=command,
                             dependencies=parameters.dependencies)

    def _run(self, config, temp):
        with gzip.open(self._in_vcf) as handle:
            with open(os.path.join(temp, "heterozygous_snps.bed"), "w") as bed:
                for line in handle:
                    if line.startswith("#"):
                        continue

                    fields = line.split("\t", 5)
                    if "," in fields[4]:
                        pos = int(fields[1])
                        bed.write("%s\t%i\t%i\n" % (fields[0], pos - 1, pos))

        CommandNode._run(self, config, temp)


class VCFFilterNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, pileup, infile, outfile, regions, dependencies=()):
        cat = factory.new("cat")
        cat.add_value("%(IN_VCF)s")
        cat.set_kwargs(IN_VCF=infile,
                       OUT_STDOUT=AtomicCmd.PIPE)

        vcffilter = factory.new("vcf_filter")
        vcffilter.add_option("--pileup", "%(IN_PILEUP)s")
        for contig in regions["HomozygousContigs"]:
            vcffilter.add_option("--homozygous-chromosome", contig)
        vcffilter.set_kwargs(IN_PILEUP=pileup,
                             IN_STDIN=cat,
                             OUT_STDOUT=AtomicCmd.PIPE)

        bgzip = AtomicCmdBuilder(["bgzip"],
                                 IN_STDIN=vcffilter,
                                 OUT_STDOUT=outfile)

        return {"commands": {"cat": cat,
                             "filter": vcffilter,
                             "bgzip": bgzip}}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        commands = [parameters.commands[key].finalize()
                    for key in ("cat", "filter", "bgzip")]

        description = "<VCFFilter: '%s' -> '%s'>" % (parameters.infile,
                                                     parameters.outfile)
        CommandNode.__init__(self,
                             description=description,
                             command=ParallelCmds(commands),
                             dependencies=parameters.dependencies)


class GenotypeRegionsNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, reference, infile, bedfile, outfile,
                  pileup_only=False, nbatches=1, dependencies=()):
        params = factory.new("genotype")
        params.add_value("%(IN_BAMFILE)s")
        params.add_value("%(OUT_VCFFILE)s")
        params.set_option("--nbatches", nbatches)

        if bedfile:
            params.set_option("--bedfile", "%(IN_INTERVALS)s")

        if pileup_only:
            params.set_option("--pileup-only")
            # Ignore read-groups for pileup
            params.add_option("--mpileup-argument", "-R", sep="=")

        # Reference sequence (FASTA)
        params.add_option("--mpileup-argument",
                          "-f=%s" % (reference,), sep="=")

        params.set_kwargs(IN_BAMFILE=infile,
                          IN_INTERVALS=bedfile,
                          OUT_VCFFILE=outfile,
                          CHECK_SAMTOOLS=SAMTOOLS_VERSION_0119,
                          CHECK_BCFTOOLS=BCFTOOLS_VERSION_0119)

        return {"command": params}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        command = parameters.command.finalize()
        invokation = " (%s%i thread(s))" \
            % ("pileup; " if parameters.pileup_only else "",
               parameters.nbatches)
        description = "<GenotypeRegions%s: '%s' -> '%s'>" \
            % (invokation,
               parameters.infile,
               parameters.outfile)

        CommandNode.__init__(self,
                             description=description,
                             command=command,
                             threads=parameters.nbatches,
                             dependencies=parameters.dependencies)


class BuildRegionsNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, infile, bedfile, outfile, padding, dependencies=()):
        params = factory.new("vcf_to_fasta")
        params.set_option("--padding", padding)
        params.set_option("--genotype", "%(IN_VCFFILE)s")
        params.set_option("--intervals", "%(IN_INTERVALS)s")

        params.set_kwargs(IN_VCFFILE=infile,
                          IN_TABIX=infile + ".tbi",
                          IN_INTERVALS=bedfile,
                          OUT_STDOUT=outfile)

        return {"command": params}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        command = parameters.command.finalize()
        description = "<BuildRegions: '%s' -> '%s'>" % (parameters.infile,
                                                        parameters.outfile)
        CommandNode.__init__(self,
                             description=description,
                             command=command,
                             dependencies=parameters.dependencies)


class SampleRegionsNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, infile, bedfile, outfile, dependencies=()):
        params = factory.new("sample_pileup")
        params.set_option("--genotype", "%(IN_PILEUP)s")
        params.set_option("--intervals", "%(IN_INTERVALS)s")
        params.set_kwargs(IN_PILEUP=infile,
                          IN_INTERVALS=bedfile,
                          OUT_STDOUT=outfile)

        return {"command": params}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        command = parameters.command.finalize()

        description = "<SampleRegions: '%s' -> '%s'>" \
            % (parameters.infile, parameters.outfile)

        CommandNode.__init__(self,
                             description=description,
                             command=command,
                             dependencies=parameters.dependencies)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is herby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Disable warning for missing docstring
# pylint: disable=C0111
# Disable warning caused by "invalid" function names
# pylint: disable=C0103
# Disable warning caused by touching private member variables/functions
# TODO: Remove this / fix places touching privates
# pylint: disable=W0212
import os

from flexmock import \
    flexmock

from paleomix.common.testing import \
    with_temp_folder, \
    set_file_contents

from paleomix.nodegraph import \
    NodeGraph, \
    FileStatusCache


def test_dir():
    return os.path.dirname(__file__)


def test_file(*args):
    return os.path.join(test_dir(), "data", *args)


_DESCRIPTION = "My description of a node"
_IN_FILES = frozenset((test_file("empty_file_1"),
                       test_file("empty_file_2")))
_OUT_FILES = frozenset((test_file("missing_out_file_1"),
                        test_file("missing_out_file_2")))
_EXEC_FILES = frozenset(("ls", "sh"))
_AUX_FILES = frozenset((test_file("rCRS.fasta"),
                        test_file("rCRS.fasta.fai")))
_REQUIREMENTS = frozenset((id, str))


###############################################################################
###############################################################################
# Setup timestamps for test files

def setup_module():
    timestamps = {test_file("timestamp_a_older"): 1000190760,
                  test_file("timestamp_b_older"): 1000190760,
                  test_file("timestamp_a_younger"): 1120719000,
                  test_file("timestamp_b_younger"): 1120719000}

    for filename, timestamp in timestamps.iteritems():
        # Set atime and mtime
        os.utime(filename, (timestamp, timestamp))


###############################################################################
###############################################################################
# NodeGraph: _is_done
# TODO: Avoid testing private function, mock cache

def test_nodegraph_is_done__no_output():
    cache = FileStatusCache()
    node = flexmock(output_files=())
    assert NodeGraph._is_done(node, cache)


@with_temp_folder
def test_nodegraph_is_done__output_changes(temp_folder):
    temp_file_1 = os.path.join(temp_folder, "file_1.txt")
    temp_file_2 = os.path.join(temp_folder, "file_2.txt")
    my_node = flexmock(output_files=(temp_file_1, temp_file_2))
    assert not NodeGraph._is_done(my_node, FileStatusCache())
    set_file_contents(temp_file_1, "foo")
    assert not NodeGraph._is_done(my_node, FileStatusCache())
    set_file_contents(temp_file_2, "bar")
    assert NodeGraph._is_done(my_node, FileStatusCache())


@with_temp_folder
def test_nodegraph_is_done__subnode_not_considered(temp_folder):
    temp_file = os.path.join(temp_folder, "file.txt")
    subnode = flexmock(output_files=(temp_file,))
    my_node = flexmock(output_files=(),
                       subnodes=(subnode,))
    assert NodeGraph._is_done(my_node, FileStatusCache())


def test_nodegraph_is_outdated__no_output():
    my_node = flexmock(input_files=(),
                       output_files=())
    assert not NodeGraph._is_outdated(my_node, FileStatusCache())


def test_nodegraph_is_outdated__input_but_no_output():
    my_node = flexmock(input_files=_IN_FILES,
                       output_files=())
    assert not NodeGraph._is_outdated(my_node, FileStatusCache())


def test_nodegraph_is_outdated__output_but_no_input():
    my_node = flexmock(input_files=(),
                       output_files=_OUT_FILES)
    assert not NodeGraph._is_outdated(my_node, FileStatusCache())


def test_nodegraph_is_outdated__not_outdated():
    my_node = flexmock(input_files=(test_file("timestamp_a_older"),),
                       output_files=(test_file("timestamp_a_younger"),))
    assert not NodeGraph._is_outdated(my_node, FileStatusCache())


def test_nodegraph_is_outdated__outdated():
    my_node = flexmock(input_files=(test_file("timestamp_a_younger"),),
                       output_files=(test_file("timestamp_a_older"),))
    assert NodeGraph._is_outdated(my_node, FileStatusCache())


def test_nodegraph_is_outdated__updates():
    my_node = flexmock(input_files=(test_file("timestamp_a_older"),),
                       output_files=(test_file("timestamp_a_younger"),))
    assert not NodeGraph._is_outdated(my_node, FileStatusCache())
    my_node = flexmock(input_files=(test_file("timestamp_a_younger"),),
                       output_files=(test_file("timestamp_a_older"),))
    assert NodeGraph._is_outdated(my_node, FileStatusCache())



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import signal
import types


def from_str(value):
    if not isinstance(value, types.StringTypes):
        raise TypeError("'from_str' takes strings, not %r" % (value.__class__.__name__,))
    return FROM_STRING[value]


def to_str(value):
    if not isinstance(value, (types.IntType, types.LongType)):
        raise TypeError("'from_str' takes strings, not %r" % (value.__class__.__name__,))
    return FROM_SIGNAL[value]


def _get_signals():
    signals = {}
    for key in dir(signal):
        if key.startswith("SIG") and not key.startswith("SIG_"):
            signals[getattr(signal, key)] = key

    # The following signals may have synonyms, so specify which to use.
    # For example, SIGIOT is a synonym of SIGABTR on OSX.
    for key in ("SIGABRT", "SIGCHLD", "SIGIO"):
        value = getattr(signal, key)
        signals[value] = key

    return signals

FROM_SIGNAL = _get_signals()
FROM_STRING = dict(zip(FROM_SIGNAL.values(), FROM_SIGNAL.keys()))



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

import paleomix.common.fileutils as \
    fileutils
from paleomix.node import \
    CommandNode
from paleomix.atomiccmd.command import \
    AtomicCmd
from paleomix.atomiccmd.builder import \
    AtomicJavaCmdBuilder
from paleomix.atomiccmd.sets import \
    ParallelCmds
from paleomix.common.fileutils import \
    swap_ext, \
    describe_files
from paleomix.common.utilities import \
    safe_coerce_to_tuple
from paleomix.nodes.samtools import \
    SAMTOOLS_VERSION
from paleomix.nodes.bwa import \
    _get_max_threads

import paleomix.common.versions as versions


def _get_gatk_version_check(config):
    """Returns a version-check object for the "GenomeAnalysisTK.jar" located at
    config.jar_root; for now, this check only serves to verify that the JAR can
    be executed, which may not be the case if the JRE is outdated.
    """
    jar_file = os.path.join(config.jar_root, "GenomeAnalysisTK.jar")
    if jar_file not in _GATK_VERSION:
        params = AtomicJavaCmdBuilder(jar_file,
                                      temp_root=config.temp_root,
                                      jre_options=config.jre_options)
        params.add_value("--version")

        # Any version is fine; for now just catch old JREs
        requirement = versions.Requirement(call=params.finalized_call,
                                           name="GenomeAnalysisTK",
                                           search=r"^(\d+)\.(\d+)",
                                           checks=versions.Any())
        _GATK_VERSION[jar_file] = requirement
    return _GATK_VERSION[jar_file]
_GATK_VERSION = {}


class GATKIndelTrainerNode(CommandNode):
    def __init__(self, config, reference, infiles, outfile,
                 threads=1, dependencies=()):
        threads = _get_max_threads(reference, threads)
        infiles = safe_coerce_to_tuple(infiles)
        jar_file = os.path.join(config.jar_root, "GenomeAnalysisTK.jar")
        command = AtomicJavaCmdBuilder(jar_file,
                                       jre_options=config.jre_options)
        command.set_option("-T", "RealignerTargetCreator")
        command.set_option("-R", "%(IN_REFERENCE)s")
        command.set_option("-o", "%(OUT_INTERVALS)s")
        command.set_option("-nt", threads)

        _set_input_files(command, infiles)
        command.set_kwargs(IN_REFERENCE=reference,
                           IN_REF_DICT=fileutils.swap_ext(reference, ".dict"),
                           OUT_INTERVALS=outfile,
                           CHECK_GATK=_get_gatk_version_check(config))

        description = "<GATK Indel Realigner (training): %s -> %r>" \
            % (describe_files(infiles), outfile)
        CommandNode.__init__(self,
                             threads=threads,
                             description=description,
                             command=command.finalize(),
                             dependencies=dependencies)


class GATKIndelRealignerNode(CommandNode):
    def __init__(self, config, reference, intervals, infiles, outfile,
                 dependencies=()):
        self._basename = os.path.basename(outfile)

        infiles = safe_coerce_to_tuple(infiles)
        jar_file = os.path.join(config.jar_root, "GenomeAnalysisTK.jar")
        command = AtomicJavaCmdBuilder(jar_file,
                                       jre_options=config.jre_options)
        command.set_option("-T", "IndelRealigner")
        command.set_option("-R", "%(IN_REFERENCE)s")
        command.set_option("-targetIntervals", "%(IN_INTERVALS)s")
        command.set_option("-o", "%(OUT_BAMFILE)s")
        command.set_option("--bam_compression", 0)
        command.set_option("--disable_bam_indexing")
        _set_input_files(command, infiles)

        command.set_kwargs(IN_REFERENCE=reference,
                           IN_REF_DICT=fileutils.swap_ext(reference, ".dict"),
                           IN_INTERVALS=intervals,
                           OUT_BAMFILE=outfile,
                           CHECK_GATK=_get_gatk_version_check(config))

        calmd = AtomicCmd(["samtools", "calmd", "-b",
                           "%(TEMP_IN_BAM)s", "%(IN_REF)s"],
                          TEMP_IN_BAM=self._basename,
                          IN_REF=reference,
                          TEMP_OUT_STDOUT=self._basename + ".calmd",
                          CHECK_VERSION=SAMTOOLS_VERSION)

        description = "<GATK Indel Realigner (aligning): %s -> %r>" \
            % (describe_files(infiles), outfile)
        CommandNode.__init__(self,
                             description=description,
                             command=ParallelCmds([command.finalize(), calmd]),
                             dependencies=dependencies)

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)
        os.mkfifo(os.path.join(temp, self._basename))

    def _teardown(self, config, temp):
        os.rename(os.path.join(temp, self._basename) + ".calmd",
                  os.path.join(temp, self._basename))

        CommandNode._teardown(self, config, temp)


def _set_input_files(command, input_files):
    keys = {}
    for (index, filename) in enumerate(input_files):
        command.add_option("-I", "%%(IN_BAMFILE_%02i)s" % index)
        keys["IN_BAMFILE_%02i" % index] = filename
        keys["IN_BAIFILE_%02i" % index] = swap_ext(filename, ".bai")

    command.set_kwargs(**keys)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.node import CommandNode
from paleomix.atomiccmd.command import AtomicCmd
from paleomix.atomiccmd.sets import SequentialCmds

from paleomix.common.fileutils import reroot_path, swap_ext
import paleomix.common.versions as versions


_VERSION_REGEX = r"Version: (\d+)\.(\d+)(?:\.(\d+))?"

# v0.2.0 was the pre-release version of v1.0, and lacks required features
_COMMON_CHECK = versions.Or(versions.EQ(0, 1, 19),
                            versions.GE(1, 0, 0))

SAMTOOLS_VERSION = versions.Requirement(call=("samtools",),
                                        search=_VERSION_REGEX,
                                        checks=_COMMON_CHECK)

SAMTOOLS_VERSION_0119 = versions.Requirement(call=("samtools",),
                                             search=_VERSION_REGEX,
                                             checks=versions.EQ(0, 1, 19))

BCFTOOLS_VERSION_0119 \
    = versions.Requirement(call=("bcftools",),
                           search=_VERSION_REGEX,
                           checks=versions.EQ(0, 1, 19))

TABIX_VERSION = versions.Requirement(call=("tabix",),
                                     search=_VERSION_REGEX,
                                     checks=versions.GE(0, 2, 5))


class TabixIndexNode(CommandNode):
    """Tabix indexes a BGZip compressed VCF or pileup file.

    The class currently supports the following presets:
        - vcf -- BGZipped VCF file.
        - pileup -- BGZipped pileup (non-binary) as produced by 'mpileup'.
    """

    def __init__(self, infile, preset="vcf", dependencies=()):
        assert infile.lower().endswith(".bgz")
        if preset == "pileup":
            call = ["tabix", "-s", 1, "-b", 2, "-e", 2]
        elif preset == "vcf":
            call = ["tabix", "-p", preset]
        else:
            assert False, "Unxpected preset: %r" % preset

        self._infile = infile
        cmd_tabix = AtomicCmd(call + ["%(TEMP_IN_VCFFILE)s"],
                              TEMP_IN_VCFFILE=os.path.basename(infile),
                              IN_VCFFILE=infile,
                              OUT_TBI=infile + ".tbi",
                              CHECK_TABIX=TABIX_VERSION)

        CommandNode.__init__(self,
                             description="<TabixIndex (%s): '%s'>" % (preset,
                                                                      infile,),
                             command=cmd_tabix,
                             dependencies=dependencies)

    def _setup(self, config, temp):
        """See CommandNode._setup."""
        infile = os.path.abspath(self._infile)
        outfile = reroot_path(temp, self._infile)
        os.symlink(infile, outfile)

        CommandNode._setup(self, config, temp)

    def _teardown(self, config, temp):
        """See CommandNode._teardown."""
        os.remove(reroot_path(temp, self._infile))

        CommandNode._teardown(self, config, temp)


class FastaIndexNode(CommandNode):
    """Indexed a FASTA file using 'samtools faidx'."""

    def __init__(self, infile, dependencies=()):
        self._infile = infile
        cmd_faidx = AtomicCmd(["samtools", "faidx", "%(TEMP_IN_FASTA)s"],
                              TEMP_IN_FASTA=os.path.basename(infile),
                              IN_FASTA=infile,
                              OUT_TBI=infile + ".fai",
                              CHECK_SAM=SAMTOOLS_VERSION)

        CommandNode.__init__(self,
                             description="<FastaIndex: '%s'>" % (infile,),
                             command=cmd_faidx,
                             dependencies=dependencies)

    def _setup(self, config, temp):
        """See CommandNode._setup."""
        infile = os.path.abspath(self._infile)
        outfile = reroot_path(temp, self._infile)
        os.symlink(infile, outfile)

        CommandNode._setup(self, config, temp)

    def _teardown(self, config, temp):
        """See CommandNode._teardown."""
        os.remove(reroot_path(temp, self._infile))

        CommandNode._teardown(self, config, temp)


class BAMIndexNode(CommandNode):
    """Indexed a BAM file using 'samtools index'."""

    def __init__(self, infile, dependencies=()):
        basename = os.path.basename(infile)

        cmd_link = AtomicCmd(["ln", "-s", "%(IN_BAM)s", "%(TEMP_OUT_BAM)s"],
                             IN_BAM=infile,
                             TEMP_OUT_BAM=basename,
                             set_cwd=True)

        cmd_index = AtomicCmd(["samtools", "index", "%(TEMP_IN_BAM)s"],
                              TEMP_IN_BAM=basename,
                              CHECK_SAM=SAMTOOLS_VERSION)

        cmd_rename = AtomicCmd(["mv", "%(TEMP_IN_BAM)s", "%(OUT_BAM)s"],
                               TEMP_IN_BAM=basename + ".bai",
                               OUT_BAM=swap_ext(infile, ".bai"))

        commands = SequentialCmds((cmd_link, cmd_index, cmd_rename))

        CommandNode.__init__(self,
                             description="<BAMIndex: '%s'>" % (infile,),
                             command=commands,
                             dependencies=dependencies)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import subprocess

from nose.tools import \
     assert_in, \
     assert_equal

import paleomix.tools.factory as factory


###############################################################################
###############################################################################

class ProcError(RuntimeError):
    pass


def check_run(call, *args, **kwargs):
    devnull = os.open(os.devnull, os.O_RDONLY)
    kwargs.setdefault("stdin", devnull)
    kwargs.setdefault("close_fds", True)
    kwargs["stdout"] = subprocess.PIPE
    kwargs["stderr"] = subprocess.PIPE

    returncode = kwargs.pop("expected_returncode", 0)

    proc = subprocess.Popen(call, *args, **kwargs)
    os.close(devnull)

    stdout, stderr = proc.communicate()
    if proc.returncode != returncode:
        raise ProcError("Command returned %i: %r:\nSTDOUT: %r\nSTDERR: %r"
                        % (proc.returncode, call, stdout, stderr))

    return stdout, stderr


# Simple test of the paleomxi command
def test_paleomix_command():
    stdout, stderr = check_run(["paleomix"])

    assert_equal("", stdout)
    assert_in("PALEOMIX - pipelines and tools for NGS data analyses.", stderr)


# Simple test that all commands can be executed
def test_factory__commands():
    def _do_test_factory__commands(command, expected):
        cmd = factory.new(command)
        call = cmd.finalized_call
        if command in ("bam_pipeline", "trim_pipeline"):
            call.append("run")

        stdout, stderr = check_run(call + ["--help"])

        assert_equal(expected, stdout.split("\n")[0])
        assert_equal("", stderr)

    commands = (("bam_pipeline",    "Usage: paleomix bam_pipeline <command> [options] [makefiles]"),
                ("trim_pipeline",   "Usage: paleomix trim_pipeline <command> [options] [makefiles]"),
                ("phylo_pipeline",  "Usage: paleomix phylo_pipeline <command> [options] [makefiles]"),
                ("cleanup",         "usage: paleomix cleanup --temp-prefix prefix --fasta reference.fasta < in.sam"),
                ("coverage",        "usage: paleomix coverage [options] sorted.bam [out.coverage]"),
                ("depths",          "usage: paleomix depths [options] sorted.bam [out.depths]"),
                ("duphist",         "usage: paleomix duphist sorted.bam > out.histogram"),
                ("rmdup_collapsed", "usage: paleomix rmdup_collapsed [options] < sorted.bam > out.bam"),
                ("genotype",        "usage: paleomix genotype [options] sorted.bam out.vcf.bgz"),
                ("gtf_to_bed",      "usage: paleomix gtf_to_bed [options] in.gtf out_prefix [in.scaffolds]"),
                ("sample_pileup",   "usage: paleomix sample_pileup [options] --genotype in.vcf --intervals in.bed > out.fasta"),
                ("vcf_filter",      "Usage: paleomix vcf_filter [options] [in1.vcf, ...]"),
                ("vcf_to_fasta",    "usage: paleomix vcf_to_fasta [options] --genotype in.vcf --intervals in.bed"),
                ("cat",             "usage: paleomix cat [options] [in|in.gz|in.bz2, ...]"))

    for command, expected in commands:
        yield _do_test_factory__commands, command, expected


# Simple test of aliased commands
def test_factory__command_alias():
    def _do_test_factory__command_alias(alias, command):
        alias, command = [alias], [command]
        if alias == ["bam_pipeline"] or alias == ["trim_pipeline"]:
            alias.append("run")
            command.append("run")

        stdout_1, stderr_1 = check_run(alias + ["--help"])
        stdout_2, stderr_2 = check_run(["paleomix"] + command + ["--help"])

        assert_equal(stderr_1, stderr_1)
        assert_equal(stderr_2, stderr_2)

    commands = (("bam_pipeline", "bam_pipeline"),
                ("bam_rmdup_collapsed", "rmdup_collapsed"),
                ("conv_gtf_to_bed", "gtf_to_bed"),
                ("phylo_pipeline", "phylo_pipeline"),
                ("trim_pipeline", "trim_pipeline"))

    for alias, command in commands:
        yield _do_test_factory__command_alias, alias, command



#!/usr/bin/python
#
# Copyright (c) 2013 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import time
import logging
import optparse

import paleomix
import paleomix.pipeline
import paleomix.ui
import paleomix.yaml
import paleomix.logger

import paleomix.common.console as console
import paleomix.tools.bam_pipeline.makefile as makefile
import paleomix.tools.cnv_pipeline.steps.preprocess as step_pre

from paleomix.config import \
    ConfigError, \
    PerHostValue, \
    PerHostConfig

from paleomix.ui import \
    print_err


_DESCRIPTION = \
    "Commands:\n" \
    "  -- %prog help        -- Display this message.\n" \
    "  -- %prog preprocess  -- TODO.\n" \
    "  -- %prog run [...]   -- TODO.\n"


class CustomHelpFormatter(optparse.IndentedHelpFormatter):
    def format_description(self, description):
        return description or ""


def _run_config_parser(argv):
    per_host_cfg = PerHostConfig("phylo_pipeline")
    usage_str = "%prog <command> [options] [makefiles]"
    version_str = "%%prog v%s" % (paleomix.__version__,)
    parser = optparse.OptionParser(usage=usage_str,
                                   version=version_str)
    parser.formatter = CustomHelpFormatter()
    parser.formatter.set_parser(parser)
    parser.description = _DESCRIPTION

    paleomix.ui.add_optiongroup(parser,
                                ui_default=PerHostValue("running"),
                                color_default=PerHostValue("on"))
    paleomix.logger.add_optiongroup(parser, default=PerHostValue("warning"))

    group = optparse.OptionGroup(parser, "Scheduling")
    group.add_option("--max-threads", default=per_host_cfg.max_threads,
                     type=int,
                     help="Maximum total number of threads to use [%default]")
    group.add_option("--dry-run", default=False, action="store_true",
                     help="If passed, only a dry-run in performed, the "
                          "dependency tree is printed, and no tasks are "
                          "executed.")
    parser.add_option_group(group)

    group = optparse.OptionGroup(parser, "Required paths")
    group.add_option("--temp-root", default=per_host_cfg.temp_root,
                     help="Location for temporary files [%default]")
    group.add_option("--destination", default="./results",
                     help="The destination for results [%default]")
    group.add_option("--mask-dir", default="./mask_dir",
                     help="The destination for results [%default]")
    parser.add_option_group(group)

    group = optparse.OptionGroup(parser, "Files and executables")
    group.add_option("--list-executables", action="store_true", default=False,
                     help="List all executables required by the pipeline, "
                          "with version requirements (if any).")
    group.add_option("--to-dot-file", dest="dot_file",
                     help="Write dependency tree to the specified dot-file.")
    parser.add_option_group(group)

    return per_host_cfg.parse_args(parser, argv)


def main(argv):
    try:
        config, args = _run_config_parser(argv)
        valid = False
        # Command, genome, makefiles
        valid |= (len(args) >= 2) and (args[0] == "preprocess")
        # Command, genome, makefiles
        valid |= (len(args) >= 2) and (args[0] == "analyse")

        if not valid:
            description = _DESCRIPTION.replace("%prog", "cnv_pipeline").strip()
            console.print_info("CNV Pipeline v%s\n" % (paleomix.__version__,))
            console.print_info(description)
            return 1

        paleomix.ui.set_ui_colors(config.ui_colors)
    except ConfigError, error:
        print_err(error)
        return 1

    if not os.path.exists(config.temp_root):
        try:
            os.makedirs(config.temp_root)
        except OSError, error:
            print_err("ERROR: Could not create temp root:\n\t%s" % (error,))
            return 1

    if not os.access(config.temp_root, os.R_OK | os.W_OK | os.X_OK):
        print_err("ERROR: Insufficient permissions for temp root: '%s'"
                  % (config.temp_root,))
        return 1

    logfile_template = time.strftime("phylo_pipeline.%Y%m%d_%H%M%S_%%02i.log")
    paleomix.logger.initialize(config, logfile_template)
    logger = logging.getLogger(__name__)

    makefiles = makefile.read_makefiles(config, args[1:])

    pipeline = paleomix.pipeline.Pypeline(config)
    if args[0] == "preprocess":
        for mkfile in makefiles:
            pipeline.add_nodes(step_pre.build_pipeline(config, mkfile))

    if config.list_executables:
        logger.info("Printing required executables ...")
        pipeline.print_required_executables()
        return 0
    elif config.dot_file:
        logger.info("Writing dependency graph to %r ...", config.dot_file)
        if not pipeline.to_dot(config.dot_file):
            return 1
        return 0

    if not pipeline.run(max_threads=config.max_threads,
                        dry_run=config.dry_run,
                        progress_ui=config.progress_ui):
        return 1
    return 0



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import sys
import argparse
import collections

import pysam

from paleomix.common.fileutils import \
    swap_ext
from paleomix.common.bedtools import \
    sort_bed_by_bamfile, \
    read_bed_file


class BAMStatsError(RuntimeError):
    pass


def collect_readgroups(args, handle):
    if args.ignore_readgroups:
        return {None: {"SM": "<NA>", "LB": "<NA>"}}

    readgroups = {}
    for readgroup in handle.header.get("RG", ()):
        key_id = readgroup["ID"]
        sample = readgroup["SM"]
        library = readgroup["LB"]

        readgroups[key_id] = {"SM": sample, "LB": library}
    return readgroups


def collect_references(args, handle):
    if args.regions:
        lengths = collections.defaultdict(int)
        for region in args.regions:
            lengths[region.name] += region.end - region.start

        lengths = dict(lengths)
    elif handle.nreferences <= args.max_contigs:
        lengths = dict(zip(handle.references, handle.lengths))
    else:
        lengths = {"<Genome>": sum(handle.lengths)}

    return lengths


def collect_bed_regions(filename):
    regions = []
    for record in read_bed_file(filename):
        if len(record) < 4:
            record.name = "%s*" % (record.contig,)

        regions.append(record)

    return regions


def parse_arguments(argv, ext):
    prog = "paleomix %s" % (ext.strip("."),)
    usage = "%s [options] sorted.bam [out%s]" % (prog, ext)
    parser = argparse.ArgumentParser(prog=prog, usage=usage)

    parser.add_argument("infile", metavar="BAM",
                        help="Filename of a sorted BAM file. If set to '-' "
                             "the file is read from STDIN.")
    parser.add_argument("outfile", metavar="OUTPUT", nargs='?',
                        help="Filename of output table; defaults to name of "
                             "the input BAM with a '%s' extension. If "
                             "set to '-' the table is printed to STDOUT."
                             % (ext,))
    parser.add_argument("--target-name", default=None, metavar="NAME",
                        help="Name used for 'Target' column; defaults to the "
                             "filename of the BAM file.")
    parser.add_argument("--regions-file", default=None, dest="regions_fpath",
                        help="BED file containing regions of interest; %s "
                             "is calculated only for these grouping by the "
                             "name used in the BED file, or the contig name "
                             "if no name has been specified for a record."
                             % (ext.strip("."),))
    parser.add_argument('--max-contigs', default=100, type=int,
                        help="The maximum number of contigs allowed in a BAM "
                             "file. If this number is exceeded, the entire "
                             "set of contigs is aggregated into one pseudo-"
                             "contig named '<Genome>'. This is done to "
                             "limit table sizes [default: %(default)s]")
    parser.add_argument('--ignore-readgroups',
                        default=False, action="store_true",
                        help="Ignore readgroup information in reads, and only "
                             "provide aggregated statistics; this is required "
                             "if readgroup information is missing or partial "
                             "[default: %(default)s]")
    parser.add_argument('--overwrite-output',
                        default=False, action="store_true",
                        help="Overwrite output file if it it exists; by "
                             "default, the script will terminate if the file "
                             "already exists.")

    args = parser.parse_args(argv)
    if not args.outfile:
        args.outfile = swap_ext(args.infile, ext)

    if args.ignore_readgroups:
        args.get_readgroup_func = _get_readgroup_ignored
    else:
        args.get_readgroup_func = _get_readgroup

    if not args.target_name:
        if args.infile == "-":
            args.target_name = "<STDIN>"
        else:
            args.target_name = os.path.basename(args.infile)

    if os.path.exists(args.outfile) and not args.overwrite_output:
        parser.error("Destination filename already exists (%r); use option "
                     "--overwrite-output to allow overwriting of this file."
                     % (args.outfile,))

    return args


def main_wrapper(process_func, argv, ext):
    args = parse_arguments(argv, ext)
    args.regions = None
    if args.regions_fpath:
        try:
            args.regions = collect_bed_regions(args.regions_fpath)
        except ValueError, error:
            sys.stderr.write("Failed to parse BED file %r:\n%s\n"
                             % (args.regions_fpath, error))
            return 1

    sys.stderr.write("Opening %r\n" % (args.infile,))
    with pysam.Samfile(args.infile) as handle:
        sort_bed_by_bamfile(handle, args.regions)
        return process_func(handle, args)


def _get_readgroup(record):
    try:
        return record.opt("RG")
    except KeyError:
        raise BAMStatsError("Record lacks readgroup tag ('RG'), use of "
                            "--ignore-readgroups is required to process this"
                            "file.\n    Record = %r" % (record.qname,))


def _get_readgroup_ignored(_):
    return None



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


# Quality score offsets for Phred (or similar) scores in FASTQ reads (33 or 64)
OFFSET_33 = 33
OFFSET_64 = 64
# Quality score found in both ranges that are unique to each offset,
# suggesting that the list contains mixed quality offsets, or invalid data.
OFFSET_BOTH = "BOTH"
# No quality scores in the expected range of ASCII values (33 .. 105)
OFFSET_MISSING = "MISSING"
# Quality scores are in the ASCII range 59 .. 74, which could signify
# low-quality reads with offset 64, or high-quality reads with offset 33
OFFSET_AMBIGIOUS = "AMBIGIOUS"


def classify_quality_strings(quality_strings):
    """Takes a sequence of quality strings from FASTQ"""
    counts = [0] * 256
    for read in quality_strings:
        for char in read:
            counts[ord(char)] += 1

    return _identify_format(counts)


def _identify_format(counts):
    """Given a list representing counts of ASCII characters found in one or
    more FASTQ quality strins, this function attempts to identify the offset
    used to encode the quality scores.

    The following constants may be returned:
      - OFFSET_33: Offset identified as being 33
      - OFFSET_64: Offset identified as being 64
      - OFFSET_BOTH: Both offset 33 and 64 found, mixed file? (error)
      - OFFSET_MISSING: No quality scores found, wrong file? (error)
      - OFFSET_AMBIGIOUS: Qualities could be either offset. (warning)
    """
    # The range of scores that can unambigiously be identified
    # as belonging to Phred scores with offset 33 or 64. Scores
    # in between could potentially signify either offset
    # See e.g. http://en.wikipedia.org/wiki/FASTQ_format#Encoding
    has_offset_33_scores = any(counts[33:59])
    has_ambigious_scores = any(counts[59:75])
    has_offset_64_scores = any(counts[75:105])

    if has_offset_33_scores:
        if has_offset_64_scores:
            return OFFSET_BOTH
        return OFFSET_33
    elif has_offset_64_scores:
        return OFFSET_64
    elif has_ambigious_scores:
        return OFFSET_AMBIGIOUS
    return OFFSET_MISSING



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import sys
import shutil
import tempfile

import nose
from nose.tools import assert_equal

from paleomix.common.fileutils import \
     make_dirs


def assert_list_equal(iter_a, iter_b):
    """Compare two values, after first converting them to lists.
    This ensures that lazily generated results can be compared."""
    list_a = list(iter_a)
    list_b = list(iter_b)

    assert_equal(list_a, list_b)


def with_temp_folder(func):
    """Decorator for unit-tests:
    Creates a unique temporary folder before running 'func'. The
    function is is assumed to take at least one parameter, the first
    of which is assumed to represent the temporary folder."""
    temp_root = os.path.join(tempfile.gettempdir(), os.getlogin())
    make_dirs(temp_root) # Ensure that this subdirectory exists

    @nose.tools.istest
    def _wrapper(*args, **kwargs):
        try:
            temp_folder = None
            temp_folder = tempfile.mkdtemp(dir    = temp_root,
                                           prefix = "paleomix_unit")
            func(temp_folder, *args, **kwargs)
        finally:
            if temp_folder:
                shutil.rmtree(temp_folder)
    _wrapper.__name__ = func.__name__ + "__wrapped_by_with_temp_folder"
    return _wrapper


class Monkeypatch:
    """Replaces a function/object in a module with the specified wrapper
     upon entry, reverting the change upon exit from the with statement.
     A full path to the given function is required, for example
       'os.path.join'."""
    def __init__(self, path, wrapper):
        self.wrapper = wrapper

        parts = path.split(".")
        assert len(parts) > 1
        self.module, self.object  = None, sys.modules[parts[0]]
        for path_cmp in parts[1:]:
            self.module, self.object = self.object, getattr(self.object, path_cmp)
        self.name = parts[-1]

    def __enter__(self):
        setattr(self.module, self.name, self.wrapper)
        return self

    def __exit__(self, _type, _value, _traceback):
        setattr(self.module, self.name, self.object)


class SetWorkingDirectory:
    """Sets the current working directory upon entry to that specified,
    in the constructor upon entry, and reverts to the previously used
    directory upon exiting a with statement."""
    def __init__(self, path):
        self._old_cwd = None
        self._new_cwd = path

    def __enter__(self):
        self._old_cwd = os.getcwd()
        os.chdir(self._new_cwd)

    def __exit__(self, _type, _value, _traceback):
        os.chdir(self._old_cwd)


def set_file_contents(fname, contents):
    with open(fname, "w") as handle:
        handle.write(contents)


def get_file_contents(fname):
    with open(fname) as handle:
        return handle.read()



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Required due to use of NotImplementedError in setattr:
# pylint: disable=R0921
import re

from paleomix.common.utilities import \
     safe_coerce_to_tuple, \
     Immutable, \
     TotallyOrdered
from paleomix.common.formats._graph import \
     GraphError, \
     _Graph


class NewickError(GraphError):
    pass


class NewickParseError(NewickError):
    """Exception raised if errors occur during parsing
    of Newick strings."""
    pass


class Newick(TotallyOrdered, Immutable):
    """Immutable object representing a Newick node.

    Nodes are classified as either internal nodes (have children),
    or leaf nodes (does not have children). A node MUST either have
    1 or more child-nodes, or have a name and/or a length. This is to
    ensure that nodes can be represented in an unambigious manner
    using the Newick format.

    No assumptions are made about the type of the 'name' and the 'length'
    properties when simply parsing the tree, and these are simply converted
    into strings when the Newick string is generated. However, additional
    contraints apply when unrooting/rerooting trees (see below). """

    def __init__(self, name = None, length = None, children = None):
        """See class documentation for constraints."""

        name = name or None
        length = length or None
        children = tuple(children or ())
        nw_hash = hash((name, length, children))
        Immutable.__init__(self,
                           name     = name,
                           length   = length,
                           children = children,
                           _hash    = nw_hash)

        if not (self.children or self.name or self.length):
            raise NewickError("Leaf nodes MUST have either a name or a length")

        # Ensure that these values are hashable
        hash(self.name)
        hash(self.length)

        weight = 0
        for child in self.children:
            if not isinstance(child, Newick):
                raise TypeError("Child nodes must be Newick nodes")
            weight += 1
        object.__setattr__(self, "_weight", weight)


    @property
    def is_leaf(self):
        """Returns true if the node is a leaf (has no children)."""
        return not self.children


    def get_leaf_nodes(self):
        """Returns iterable for leaf-nodes accessible from this node."""
        if not self.is_leaf:
            for child in self.children:
                for leaf in child.get_leaf_nodes():
                    yield leaf
        else:
            yield self


    def get_leaf_names(self):
        for node in self.get_leaf_nodes():
            yield node.name


    def reroot_on_taxa(self, taxa):
        """Returns the Newick tree from this node, but rooted on the midpoint
        of the branch leading to one or more taxa. Note that the taxa are not
        required to form a clade. If the taxa do not form a monophyletic clade,
        then the outgroup will include more taxa than those passed to the
        function."""
        return _NewickGraph(self).reroot_on_taxa(taxa)


    def reroot_on_midpoint(self):
        """Returns the newick tree from this node, but rooted on the midpoint
        of the tree. That is to say that a root node is added at the exact
        midpoint of the longest path in the unrooted tree. If this midpoint
        lies at an existing internal node, then this node is made the root.

        Note that the sorting of nodes is not preserved, and that any
        uninformative nodes (lacking name/length, while connecting two
        other nodes, e.g. the old root) are spliced out.

        All nodes must have a length of zero or greater (no missing values
        are allowed), but note that rerooting behavior around nodes with
        length zero may yield unexpected results."""
        if len(list(self.get_leaf_nodes())) < 2:
            return self # No meaningful way to reroot such trees

        return _NewickGraph(self).reroot_on_midpoint()


    def add_support(self, bootstraps, fmt = "{Support}"):
        """Adds support values to the current tree, based on a set of trees containing
        the same taxa. It is assumed that the support trees represent unrooted or
        arbitarily rooted trees, and no weight is given to the rooted topology of these
        trees.

        The main tree should itself be rooted, and the the toplogy and ordering of this
        tree is preserved, with node-names updated using the formatting string 'fmt'.

        Formatting is carried out using str.format, with these fields:
          {Support}    -- The total number of trees in which a clade is supported.
          {Percentage} -- The percentage of trees in which a clade is supported (float).
          {Fraction}   -- The fraction of trees in which a clade is supported (float).

        For example, typical percentage support-values can be realized by setting 'fmt'
        to the value "{Percentage:.0f}" to produce integer values.
        """
        clade_counts = {}
        leaf_names_lst = list(self.get_leaf_names())
        leaf_names = frozenset(leaf_names_lst)
        if len(leaf_names) != len(leaf_names_lst):
            raise NewickError("Cannot add support values to trees with duplicate leaf names")

        bootstraps   = safe_coerce_to_tuple(bootstraps)
        for support_tree in bootstraps:
            support_tree_names = frozenset(support_tree.get_leaf_names())
            if leaf_names != support_tree_names:
                raise NewickError("Support tree does not contain same set of leaf nodes")

            support_graph      = _NewickGraph(support_tree)
            for clade in support_graph.get_clade_names():
                clade_counts[clade] = clade_counts.get(clade, 0) + 1

        return self._add_support(self, len(bootstraps), clade_counts, fmt)


    @classmethod
    def from_string(cls, string):
        """Parses a Newick string and returns a representation of the tree.
        See e.g. http://en.wikipedia.org/wiki/Newick_format

        Note that implicit nodes, such as (), (A,), and the like are not
        allowed, as they cannot always be represented/parsed in an unambigious
        manner. Thus all leaf nodes must have a name and/or a length."""
        tokens = _tokenize(string)
        if tokens and tokens[0] == "(":
            top_node = _parse_tokens(tokens)
        else:
            top_node = _parse_child(tokens)

        if tokens != [";"]:
            raise NewickParseError("Missing terminating semi-colon")

        return top_node


    def __lt__(self, other):
        """See TotallyOrdered"""
        if not isinstance(other, Newick):
            return NotImplemented

        # pylint: disable=W0212
        return (-self._weight, self.name, self.length, self.children) \
          < (-other._weight, other.name, other.length, other.children)


    def __hash__(self):
        """Hashing function, see 'hash'."""
        return self._hash


    def __repr__(self):
        """Representation corresponds to the Newick string for the (sub)tree,
        which can be parsed by 'from_string'."""
        return "%s;" % (self._to_str(),)


    def _to_str(self):
        fields = []
        if self.children:
            fields.append("(")
            for child in self.children:
                fields.append(child._to_str()) # pylint: disable=W0212
                fields.append(",")
            fields.pop()
            fields.append(")")
        if self.name is not None:
            fields.append(str(self.name))
        if self.length is not None:
            fields.append(":")
            fields.append(str(self.length))
        return "".join(fields)


    def _add_support(self, node, total, clade_counts, fmt):
        """Recursively annotates a subtree with support values,
        excepting leaf nodes (where the name is preserved) and
        the root node (where the name is cleared)."""
        if node.is_leaf:
            return node

        clade   = frozenset(leaf.name for leaf in node.get_leaf_nodes())
        support = clade_counts.get(clade, 0)
        name = fmt.format(Support    = support,
                          Percentage = (support * 100.0) / (total or 1),
                          Fraction   = (support * 1.0) / (total or 1))

        children = []
        for child in node.children:
            children.append(self._add_support(child, total, clade_counts, fmt))

        return Newick(name     = (None if (node is self) else name),
                      length   = node.length,
                      children = children)





################################################################################
################################################################################
## Functions related to NEWICK parsing

_TOKENIZER = re.compile("([():,;])")
_NODE_KEYS = frozenset(("name", "length", "children"))


def _tokenize(string):
    result = []
    for field in _TOKENIZER.split(string):
        field = field.strip()
        if field:
            result.append(field)
    return result


def _parse_tokens(tokens):
    assert tokens and tokens[0] == "("

    tokens.pop(0)
    child, children = None, []
    while tokens and (tokens[0] not in ");"):
        if tokens[0] == ",":
            children.append(child)
            tokens.pop(0)
        child = _parse_child(tokens)
    children.append(child)

    if any(child is None for child in children):
        raise NewickParseError("Implicit leaf nodes (no name OR length) are not allowed")
    elif not tokens or (tokens[0] != ")"):
        raise NewickParseError("Malformed Newick string, contains unbalanced parantheses")
    tokens.pop(0)

    return _parse_child(tokens, children = children)


def _parse_child(tokens, children = None):
    if tokens and tokens[0] == "(":
        return _parse_tokens(tokens)

    name, length = None, None
    while tokens and (tokens[0] not in ",);"):
        if (tokens[0] == ":"):
            if length is not None:
                raise NewickParseError("Node has multiple length values")
            tokens.pop(0)
            if tokens[0] in ",);":
                raise NewickParseError("Missing length value")
            length = tokens.pop(0).strip()
        else:
            name = tokens.pop(0).strip()

    if not (name or length or children):
        raise NewickParseError("Parsing of implicit nodes not supported")

    return Newick(name     = name,
                  length   = length,
                  children = children)



################################################################################
################################################################################
## Class related to tree manipulations

class _NewickGraph(_Graph):
    def __init__(self, node):
        _Graph.__init__(self)
        self._collect_names_and_blengths(node)
        self.prune_uninformative_nodes()


    def _collect_names_and_blengths(self, c_node):
        c_node_id = id(c_node)

        self.set_name(c_node_id,  c_node.name)
        for child in c_node.children:
            child_id = id(child)
            self.add_connection(c_node_id, child_id, child.length)
            self._collect_names_and_blengths(child)


    def rebuild_tree(self, parent_id, node_id):
        """Rebuilds a newick tree starting at a node with id
        'node_id' and a parent with id 'parent_id' (or the
        same value as 'node_id' if a root node)."""

        children = []
        for child_id in self.connections[node_id]:
            if child_id != parent_id:
                children.append(self.rebuild_tree(node_id, child_id))
        children.sort()

        blength = self.connections.get(parent_id).get(node_id)
        if isinstance(blength, float):
            blength = repr(blength)

        return Newick(name     = self.names.get(node_id),
                      length   = blength,
                      children = children)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import collections

from paleomix.common.fileutils import \
    swap_ext

from paleomix.nodes.commands import \
    CoverageNode, \
    MergeCoverageNode, \
    DepthHistogramNode
from paleomix.tools.bam_pipeline.parts.summary import \
    SummaryTableNode


def add_statistics_nodes(config, makefile, target):
    features = makefile["Options"]["Features"]

    nodes = []
    if features["Depths"]:
        nodes.extend(_build_depth(config, target))

    if features["Summary"] or features["Coverage"]:
        make_summary = features["Summary"]
        coverage = _build_coverage(config, target, make_summary)
        if make_summary:
            summary_node = _build_summary_node(config, makefile,
                                               target, coverage)
            nodes.append(summary_node)
        elif features["Coverage"]:
            nodes.extend(coverage["Nodes"])

    target.nodes.extend(nodes)


def _build_summary_node(config, makefile, target, coverage):
    coverage_by_label = _build_coverage_nodes(config, target, use_label=True)

    return SummaryTableNode(config=config,
                            makefile=makefile,
                            target=target,
                            cov_for_lanes=coverage_by_label["Lanes"],
                            cov_for_libs=coverage_by_label["Libraries"],
                            dependencies=coverage["Nodes"])


def _build_depth(config, target):
    nodes = []
    for prefix in target.prefixes:
        for (roi_name, roi_filename) in _get_roi(prefix, name_prefix="."):
            if roi_filename is not None:
                # ROIs require indexed access, and hence that the final BAM
                # (either raw or realigned) has been built. By default, the
                # the realigned BAM is used (based on lexical order).
                bam_files = tuple(sorted(prefix.bams.items()))
                input_files, dependencies = bam_files[-1]
            else:
                input_files = {}
                for sample in prefix.samples:
                    input_files.update(sample.bams)
                dependencies = input_files.values()
                input_files = input_files.keys()

            output_filename = "%s.%s%s.depths" % (target.name, prefix.name,
                                                  roi_name)
            output_fpath = os.path.join(config.destination, output_filename)

            node = DepthHistogramNode(config=config,
                                      target_name=target.name,
                                      input_files=input_files,
                                      regions_file=roi_filename,
                                      output_file=output_fpath,
                                      dependencies=dependencies)
            nodes.append(node)

    return nodes


def _aggregate_for_prefix(cov, prefix, roi_name=None, into=None):
    prefix = _get_prefix_label(prefix, roi_name)
    results = {} if into is None else into
    for (key, files_and_nodes) in cov.iteritems():
        if prefix is None or (key[0] == prefix):
            results.update(files_and_nodes)
    return results


def _build_coverage(config, target, make_summary):
    merged_nodes = []
    coverage = _build_coverage_nodes(config, target)
    for prefix in target.prefixes:
        for (roi_name, _) in _get_roi(prefix):
            label = _get_prefix_label(prefix.name, roi_name)
            if not roi_name:
                postfix = prefix.name
            else:
                postfix = "%s.%s" % (prefix.name, roi_name)

            files_and_nodes = _aggregate_for_prefix(coverage["Libraries"],
                                                    label)
            output_filename = os.path.join(config.destination,
                                           "%s.%s.coverage"
                                           % (target.name, postfix))
            merged = MergeCoverageNode(input_files=files_and_nodes.keys(),
                                       output_file=output_filename,
                                       dependencies=files_and_nodes.values())

            merged_nodes.append(merged)

    files_and_nodes = _aggregate_for_prefix(coverage["Libraries"], None)
    if make_summary:
        files_and_nodes = _aggregate_for_prefix(coverage["Lanes"], None,
                                                into=files_and_nodes)

    all_nodes = []
    all_nodes.extend(files_and_nodes.itervalues())
    all_nodes.extend(merged_nodes)

    coverage["Nodes"] = tuple(all_nodes)

    return coverage


def _build_coverage_nodes(config, target, use_label=False):
    coverage = {"Lanes": collections.defaultdict(dict),
                "Libraries": collections.defaultdict(dict)}

    cache = {}
    for prefix in target.prefixes:
        for (roi_name, roi_filename) in _get_roi(prefix):
            prefix_label = prefix.label if use_label else prefix.name
            prefix_label = _get_prefix_label(prefix_label, roi_name)

            for sample in prefix.samples:
                for library in sample.libraries:
                    key = (prefix_label, target.name,
                           sample.name, library.name)

                    for lane in library.lanes:
                        for bams in lane.bams.values():
                            bams = _build_coverage_nodes_cached(config, bams,
                                                                target.name,
                                                                roi_name,
                                                                roi_filename,
                                                                cache)

                            coverage["Lanes"][key].update(bams)

                    bams = _build_coverage_nodes_cached(config, library.bams,
                                                        target.name, roi_name,
                                                        roi_filename, cache)
                    coverage["Libraries"][key].update(bams)
    return coverage


def _build_coverage_nodes_cached(config, files_and_nodes, target_name,
                                 roi_name, roi_filename, cache):
    output_ext = ".coverage"
    if roi_name:
        output_ext = ".%s.coverage" % roi_name

    coverages = {}
    for (input_filename, node) in files_and_nodes.iteritems():
        output_filename = swap_ext(input_filename, output_ext)

        cache_key = (roi_filename, input_filename)
        if cache_key not in cache:
            cache[cache_key] = CoverageNode(config=config,
                                            input_file=input_filename,
                                            output_file=output_filename,
                                            target_name=target_name,
                                            regions_file=roi_filename,
                                            dependencies=node)

        coverages[output_filename] = cache[cache_key]
    return coverages


def _get_roi(prefix, name_prefix=""):
    roi = [("", None)]
    for (name, path) in prefix.roi.iteritems():
        roi.append((name_prefix + name, path))
    return roi


def _get_prefix_label(label, roi_name):
    if not roi_name:
        return label
    return "%s:%s" % (label, roi_name)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import stat
import errno

import nose
from nose.tools import \
    assert_is, \
    assert_in, \
    assert_equal, \
    assert_raises

import paleomix
from paleomix.common.testing import \
    with_temp_folder, \
    Monkeypatch, \
    SetWorkingDirectory, \
    set_file_contents, \
    get_file_contents

from paleomix.common.fileutils import \
    add_postfix, \
    swap_ext, \
    reroot_path, \
    create_temp_dir, \
    missing_files, \
    modified_after, \
    is_executable, \
    which_executable, \
    executable_exists, \
    missing_executables, \
    make_dirs, \
    move_file, \
    copy_file, \
    open_ro, \
    try_rmdir, \
    try_remove, \
    try_rmtree, \
    describe_files, \
    describe_paired_files


###############################################################################
###############################################################################
# Setup timestamps for test files

def test_dir():
    return os.path.dirname(os.path.dirname(__file__))


def test_file(*args):
    return os.path.join(test_dir(), "data", *args)


def setup_module():
    timestamps = {test_file("timestamp_a_older"): 1000190760,
                  test_file("timestamp_b_older"): 1000190760,
                  test_file("timestamp_a_younger"): 1120719000,
                  test_file("timestamp_b_younger"): 1120719000}

    for filename, timestamp in timestamps.iteritems():
        # Set atime and mtime
        os.utime(filename, (timestamp, timestamp))


###############################################################################
###############################################################################
# Tests for 'add_postfix'

def test_add_postfix__no_postfix():
    assert_equal(add_postfix("name.foo", ""), "name.foo")


def test_add_postfix__dot_postfix():
    assert_equal(add_postfix("name.foo", ".pf"), "name.pf.foo")


def test_add_postfix__underscore_postfix():
    assert_equal(add_postfix("name.foo", "_pf"), "name_pf.foo")


def test_add_postfix__no_ext__no_postfix():
    assert_equal(add_postfix("name", ""), "name")


def test_add_postfix__no_ext__dot_postfix():
    assert_equal(add_postfix("name", ".pf"), "name.pf")


def test_add_postfix__no_ext__underscore_postfix():
    assert_equal(add_postfix("name", "_pf"), "name_pf")


###############################################################################
###############################################################################
# Tests for 'swap_ext'

def test_swap_ext__has_ext_vs_empty_ext():
    assert_equal(swap_ext("name.foo", ""), "name")


def test_swap_ext__empty_ext_vs_empty_ext():
    assert_equal(swap_ext("name", ""), "name")


def test_swap_ext__has_ext_vs_dot_ext():
    assert_equal(swap_ext("name.foo", "."), "name")


def test_swap_ext__dot_ext_vs_dot_ext():
    assert_equal(swap_ext("name.", "."), "name")


def test_swap_ext__multiple__has_ext_vs_empty_ext():
    assert_equal(swap_ext("name.foo.bar", ""), "name.foo")


def test_swap_ext__multiple__has_ext_vs_dot_ext():
    assert_equal(swap_ext("name.foo.bar", "."), "name.foo")


def test_swap_ext__multiple__dot_ext_vs_dot_ext():
    assert_equal(swap_ext("name.foo.", "."), "name.foo")


def test_swap_ext__has_ext_vs_new_ext():
    assert_equal(swap_ext("name.foo", "bar"), "name.bar")


def test_swap_ext__has_ext_vs_new_dot_ext():
    assert_equal(swap_ext("name.foo", ".bar"), "name.bar")


def test_swap_ext__empty_ext_vs_new_ext():
    assert_equal(swap_ext("name", "bar"), "name.bar")


def test_swap_ext__dot_ext_vs_new_dot_ext():
    assert_equal(swap_ext("name", ".bar"), "name.bar")


###############################################################################
###############################################################################
# Tests for 'reroot_path'

def test_reroot_path__empty_root():
    assert_equal(reroot_path("", "/etc/apt/sources.list"), "sources.list")


def test_reroot_path__empty_path():
    assert_equal(reroot_path("/etc/apt", ""), "/etc/apt/")


def test_reroot_path__abs_abs__wo_final_dash():
    assert_equal(reroot_path("/etc/apt", "/tmp/sources.list"),
                 "/etc/apt/sources.list")


def test_reroot_path__abs_abs__w_final_dash():
    assert_equal(reroot_path("/etc/apt/", "/tmp/sources.list"),
                 "/etc/apt/sources.list")


def test_reroot_path__abs_rel__wo_final_dash():
    assert_equal(reroot_path("/etc/apt", "tmp/sources.list"),
                 "/etc/apt/sources.list")


def test_reroot_path__abs_rel__w_final_dash():
    assert_equal(reroot_path("/etc/apt/", "tmp/sources.list"),
                 "/etc/apt/sources.list")


def test_reroot_path__rel_abs__wo_final_dash():
    assert_equal(reroot_path("etc/apt", "/tmp/sources.list"),
                 "etc/apt/sources.list")


def test_reroot_path__rel_abs__w_final_dash():
    assert_equal(reroot_path("etc/apt/", "/tmp/sources.list"),
                 "etc/apt/sources.list")


def test_reroot_path__rel_rel__wo_final_dash():
    assert_equal(reroot_path("etc/apt", "tmp/sources.list"),
                 "etc/apt/sources.list")


def test_reroot_path__rel_rel__w_final_dash():
    assert_equal(reroot_path("etc/apt/", "tmp/sources.list"),
                 "etc/apt/sources.list")


###############################################################################
###############################################################################
# Tests for 'create_temp_dir'

@with_temp_folder
def test_create_temp_dir__create(temp_folder):
    tmp_dir_1 = create_temp_dir(temp_folder)
    tmp_dir_2 = create_temp_dir(temp_folder)
    assert os.path.exists(tmp_dir_1)
    assert os.path.exists(tmp_dir_2)


@with_temp_folder
def test_create_temp_dir__empty(temp_folder):
    tmp_dir = create_temp_dir(temp_folder)
    contents = os.listdir(tmp_dir)
    assert not contents


@with_temp_folder
def test_create_temp_dir__permissions(temp_folder):
    tmp_dir = create_temp_dir(temp_folder)
    stats = os.stat(tmp_dir)
    assert_equal(stats.st_mode & 0777, 0750)


@with_temp_folder
def test_create_temp_dir__creation_preempted(temp_folder):
    unwrapped, preempted_once = os.makedirs, []

    def _wrap_os_makedirs(*args, **kwargs):
        # Simulate somebody else creating the directory first
        if not preempted_once:
            unwrapped(*args, **kwargs)
            preempted_once.append(True)
        unwrapped(*args, **kwargs)

    with Monkeypatch("os.makedirs", _wrap_os_makedirs):
        assert not os.listdir(temp_folder)
        work_dir = create_temp_dir(temp_folder)
        assert os.path.exists(temp_folder)
        dirs = os.listdir(temp_folder)
        assert_equal(len(dirs), 2)
        assert_in(os.path.basename(work_dir), dirs)
        assert bool(preempted_once)


def test_create_temp_dir__permission_denied():
    def _wrap_os_makedirs(*_args, **_kwargs):
        raise OSError((errno.EACCES, "Simulated premission denied"))

    with Monkeypatch("os.makedirs", _wrap_os_makedirs):
        assert_raises(OSError, create_temp_dir, "/tmp")


###############################################################################
###############################################################################
# Tests for 'missing_files'

def test_missing_files__file_exists():
    assert_equal(missing_files([test_file("empty_file_1")]), [])


def test_missing_files__file_doesnt_exist():
    assert_equal(missing_files([test_file("missing_file_1")]),
                 [test_file("missing_file_1")])


def test_missing_files__mixed_files():
    files = [test_file("missing_file_1"),
             test_file("empty_file_1")]
    result = [test_file("missing_file_1")]

    assert_equal(missing_files(files), result)


###############################################################################
###############################################################################
# Tests for 'modified_after'

def test_modified_after__modified_after():
    assert modified_after(test_file("timestamp_a_younger"),
                          test_file("timestamp_a_older"))
    assert modified_after(test_file("timestamp_a_younger"),
                          test_file("timestamp_b_older"))
    assert modified_after(test_file("timestamp_b_younger"),
                          test_file("timestamp_a_older"))


def test_modified_after__not_modified_after():
    assert not modified_after(test_file("timestamp_a_older"),
                              test_file("timestamp_a_younger"))
    assert not modified_after(test_file("timestamp_a_older"),
                              test_file("timestamp_b_younger"))
    assert not modified_after(test_file("timestamp_b_older"),
                              test_file("timestamp_a_younger"))


def test_modified_after__same_file():
    assert not modified_after(test_file("timestamp_a_older"),
                              test_file("timestamp_a_older"))
    assert not modified_after(test_file("timestamp_a_older"),
                              test_file("timestamp_b_older"))
    assert not modified_after(test_file("timestamp_b_older"),
                              test_file("timestamp_a_older"))


###############################################################################
###############################################################################
# Tests for 'is_executable'

def test_is_executable__full_path__is_executable():
    assert is_executable("/bin/ls")


def test_is_executable__full_path__is_non_executable():
    assert not is_executable("/etc/fstab")


def test_is_executable__full_path__folder_is_non_executable():
    assert not is_executable("/etc")


def test_is_executable__rel_path__is_executable():
    assert is_executable(os.path.join(test_dir(), "run"))


def test_is_executable__rel_path__is_non_executable():
    assert not is_executable(test_file("empty_file_1"))


###############################################################################
###############################################################################
# Tests for 'which_executable'

def test_which_executable__executable():
    assert_equal("/bin/ls", which_executable("ls"))


def test_which_executable__non_executable():
    assert_is(None, which_executable("lsxxxx"))


def test_which_executable__executable__but_no_path():
    path = os.environ.pop('PATH')
    try:
        assert_is(None, which_executable("ls"))
    finally:
        os.environ['PATH'] = path


def test_which_executable__executable__but_empty_path():
    path = os.environ.pop('PATH')
    try:
        os.environ['PATH'] = ""
        assert_is(None, which_executable("ls"))
    finally:
        os.environ['PATH'] = path


def test_which_executable__executable__by_path_order_1():
    path = os.environ.pop('PATH')
    try:
        path_1 = test_dir()
        path_2 = os.path.join(os.getcwd(), path_1)

        os.environ['PATH'] = ":".join((path_1, path_2))
        assert_equal(os.path.join(path_1, "run"), which_executable("run"))
    finally:
        os.environ['PATH'] = path


def test_which_executable__executable__by_path_order_2():
    path = os.environ.pop('PATH')
    try:
        path_1 = test_dir()
        path_2 = os.path.join(os.getcwd(), path_1)

        os.environ['PATH'] = ":".join((path_2, path_1))
        assert_equal(os.path.join(path_2, "run"), which_executable("run"))
    finally:
        os.environ['PATH'] = path


###############################################################################
###############################################################################
# Tests for 'executable_exists'

def test_executable_exists__executable():
    assert executable_exists("ls")


def test_executable_exists__non_executable():
    assert not executable_exists("lsxxxx")


def test_executable_exists__full_path__is_executable():
    assert executable_exists("/bin/ls")


def test_executable_exists__full_path__is_non_executable():
    assert not executable_exists("/etc/fstab")


def test_executable_exists__rel_path__is_executable():
    assert executable_exists(os.path.join(test_dir(), "run"))


def test_executable_exists__rel_path__is_non_executable():
    assert not executable_exists(test_file("empty_file_1"))


###############################################################################
###############################################################################
# Tests for 'missing_executables'

def test_missing_executables__executable():
    assert_equal(missing_executables(["ls"]), [])


def test_missing_executables__non_executable():
    assert_equal(missing_executables(["lsxxxx"]), ["lsxxxx"])


def test_missing_executables__mixed():
    assert_equal(missing_executables(["lsxxxx", "ls"]), ["lsxxxx"])


###############################################################################
###############################################################################
# Tests for 'make_dirs'

@with_temp_folder
def test_make_dirs__create_dir(temp_folder):
    assert not os.listdir(temp_folder)
    assert make_dirs(os.path.join(temp_folder, "test123"))
    assert_equal(os.listdir(temp_folder), ["test123"])


@with_temp_folder
def test_make_dirs__return_values(temp_folder):
    assert make_dirs(os.path.join(temp_folder, "test234"))
    assert not make_dirs(os.path.join(temp_folder, "test234"))


@with_temp_folder
def test_make_dirs__subdirs_return_values(temp_folder):
    assert make_dirs(os.path.join(temp_folder, "test"))
    assert make_dirs(os.path.join(temp_folder, "test", "234"))
    assert not make_dirs(os.path.join(temp_folder, "test", "234"))


@with_temp_folder
def test_make_dirs__sub_directories(temp_folder):
    assert not os.listdir(temp_folder)
    assert make_dirs(os.path.join(temp_folder, "test", "123"))
    assert_equal(os.listdir(temp_folder), ["test"])
    assert_equal(os.listdir(os.path.join(temp_folder, "test")), ["123"])


@with_temp_folder
def test_make_dirs__permissions(temp_folder):
    work_dir = os.path.join(temp_folder, "test_1")
    assert make_dirs(work_dir, mode=0511)
    stats = os.stat(work_dir)
    assert_equal(oct(stats.st_mode & 0777), oct(0511))


@with_temp_folder
def test_make_dirs__creation_preemted(temp_folder):
    unwrapped, preempted = os.makedirs, []

    def _wrap_os_makedirs(*args, **kwargs):
        # Simulate somebody else creating the directory first
        preempted.append(True)
        unwrapped(*args, **kwargs)
        unwrapped(*args, **kwargs)

    with Monkeypatch("os.makedirs", _wrap_os_makedirs):
        work_folder = os.path.join(temp_folder, "test")
        assert not make_dirs(work_folder)
        assert os.path.exists(work_folder)
        assert_equal(os.listdir(temp_folder), ["test"])
        assert_equal(preempted, [True])


@with_temp_folder
def test_make_dirs__permission_denied(temp_folder):
    # Make temporary folder read-only
    mode = os.stat(temp_folder).st_mode
    ro_mode = mode & ~(stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH)
    os.chmod(temp_folder, ro_mode)
    # Non OEXIST errors should be re-raised:
    assert_raises(OSError, make_dirs, os.path.join(temp_folder, "foo"))


@nose.tools.raises(ValueError)
def test_make_dirs__empty_directory():
    make_dirs("")


###############################################################################
###############################################################################
# Tests for 'move_file'

@with_temp_folder
def test_move_file__simple_move(temp_folder):
    file_1 = os.path.join(temp_folder, "file_1")
    file_2 = os.path.join(temp_folder, "file_2")
    assert_equal(os.listdir(temp_folder), [])
    set_file_contents(file_1, "1")
    assert_equal(os.listdir(temp_folder), ["file_1"])
    move_file(file_1, file_2)
    assert_equal(os.listdir(temp_folder), ["file_2"])
    assert_equal(get_file_contents(file_2), "1")


@with_temp_folder
def test_move_file__simple_move_in_cwd(temp_folder):
    with SetWorkingDirectory(temp_folder):
        assert_equal(os.listdir("."), [])
        set_file_contents("file_1", "1")
        assert_equal(os.listdir("."), ["file_1"])
        move_file("file_1", "file_2")
        assert_equal(os.listdir("."), ["file_2"])
        assert_equal(get_file_contents("file_2"), "1")


@with_temp_folder
def test_move_dirs__permission_denied(temp_folder):
    dst_folder = os.path.join(temp_folder, "dst")
    file_1 = os.path.join(temp_folder, "file")
    file_2 = os.path.join(dst_folder, "file")
    set_file_contents(file_1, "1")

    # Make destination folder read-only
    assert make_dirs(os.path.join(temp_folder, "dst"))
    mode = os.stat(dst_folder).st_mode
    ro_mode = mode & ~(stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH)
    os.chmod(dst_folder, ro_mode)

    # Non ENOENT errors should be re-raised:
    assert_raises(IOError, move_file, file_1, file_2)


@with_temp_folder
def test_move_file__move_to_existing_folder(temp_folder):
    assert make_dirs(os.path.join(temp_folder, "src"))
    assert make_dirs(os.path.join(temp_folder, "dst"))
    file_1 = os.path.join(temp_folder, "src", "file_1")
    file_2 = os.path.join(temp_folder, "dst", "file_2")
    set_file_contents(file_1, "2")
    move_file(file_1, file_2)
    assert_equal(os.listdir(os.path.dirname(file_1)), [])
    assert_equal(os.listdir(os.path.dirname(file_2)), ["file_2"])
    assert_equal(get_file_contents(file_2), "2")


@with_temp_folder
def test_move_file__move_to_new_folder(temp_folder):
    assert make_dirs(os.path.join(temp_folder, "src"))
    file_1 = os.path.join(temp_folder, "src", "file_1")
    file_2 = os.path.join(temp_folder, "dst", "file_2")
    set_file_contents(file_1, "2")
    move_file(file_1, file_2)
    assert_equal(os.listdir(os.path.dirname(file_1)), [])
    assert_equal(os.listdir(os.path.dirname(file_2)), ["file_2"])
    assert_equal(get_file_contents(file_2), "2")


@with_temp_folder
def test_move_file__move_to_different_folder(temp_folder):
    with SetWorkingDirectory(temp_folder):
        set_file_contents("file_1", "3")
        move_file("file_1", "dst/file_1")
        assert_equal(os.listdir("."), ["dst"])
        assert_equal(os.listdir("dst"), ["file_1"])
        assert_equal(get_file_contents("dst/file_1"), "3")


@with_temp_folder
def test_move_file__overwrite(temp_folder):
    with SetWorkingDirectory(temp_folder):
        set_file_contents("file_1", "4")
        set_file_contents("file_2", "5")
        move_file("file_1", "file_2")
        assert_equal(os.listdir("."), ["file_2"])
        assert_equal(get_file_contents("file_2"), "4")


def test_move_file__enoent_reraised_if_not_due_to_missing_folder():
    assert_raises(IOError, move_file, "", "./dst")


###############################################################################
###############################################################################
# Tests for 'copy_file'

@with_temp_folder
def test_copy_file__simple_copy(temp_folder):
    file_1 = os.path.join(temp_folder, "file_1")
    file_2 = os.path.join(temp_folder, "file_2")
    assert_equal(os.listdir(temp_folder), [])
    set_file_contents(file_1, "1")
    assert_equal(os.listdir(temp_folder), ["file_1"])
    copy_file(file_1, file_2)
    assert_equal(set(os.listdir(temp_folder)), set(["file_1", "file_2"]))
    assert_equal(get_file_contents(file_1), "1")
    assert_equal(get_file_contents(file_2), "1")


@with_temp_folder
def test_copy_file__simple_copy_in_cwd(temp_folder):
    with SetWorkingDirectory(temp_folder):
        assert_equal(os.listdir("."), [])
        set_file_contents("file_1", "1")
        assert_equal(os.listdir("."), ["file_1"])
        copy_file("file_1", "file_2")
        assert_equal(set(os.listdir(".")), set(["file_1", "file_2"]))
        assert_equal(get_file_contents("file_1"), "1")
        assert_equal(get_file_contents("file_2"), "1")


@with_temp_folder
def test_copy_file__copy_to_existing_folder(temp_folder):
    assert make_dirs(os.path.join(temp_folder, "src"))
    assert make_dirs(os.path.join(temp_folder, "dst"))
    file_1 = os.path.join(temp_folder, "src", "file_1")
    file_2 = os.path.join(temp_folder, "dst", "file_2")
    set_file_contents(file_1, "2")
    copy_file(file_1, file_2)
    assert_equal(os.listdir(os.path.dirname(file_1)), ["file_1"])
    assert_equal(os.listdir(os.path.dirname(file_2)), ["file_2"])
    assert_equal(get_file_contents(file_1), "2")
    assert_equal(get_file_contents(file_2), "2")


@with_temp_folder
def test_copy_file__copy_to_new_folder(temp_folder):
    assert make_dirs(os.path.join(temp_folder, "src"))
    file_1 = os.path.join(temp_folder, "src", "file_1")
    file_2 = os.path.join(temp_folder, "dst", "file_2")
    set_file_contents(file_1, "2")
    copy_file(file_1, file_2)
    assert_equal(os.listdir(os.path.dirname(file_1)), ["file_1"])
    assert_equal(os.listdir(os.path.dirname(file_2)), ["file_2"])
    assert_equal(get_file_contents(file_1), "2")
    assert_equal(get_file_contents(file_2), "2")


@with_temp_folder
def test_copy_file__copy_to_different_folder(temp_folder):
    with SetWorkingDirectory(temp_folder):
        set_file_contents("file_1", "3")
        copy_file("file_1", "dst/file_1")
        assert_equal(set(os.listdir(".")), set(["file_1", "dst"]))
        assert_equal(os.listdir("dst"), ["file_1"])
        assert_equal(get_file_contents("file_1"), "3")
        assert_equal(get_file_contents("dst/file_1"), "3")


@with_temp_folder
def test_copy_file__overwrite(temp_folder):
    with SetWorkingDirectory(temp_folder):
        set_file_contents("file_1", "4")
        set_file_contents("file_2", "5")
        copy_file("file_1", "file_2")
        assert_equal(set(os.listdir(".")), set(["file_1", "file_2"]))
        assert_equal(get_file_contents("file_1"), "4")
        assert_equal(get_file_contents("file_2"), "4")


def test_copy_file__enoent_reraised_if_not_due_to_missing_folder():
    assert_raises(IOError, copy_file, "", "./dst")


###############################################################################
###############################################################################
# Tests for 'open'


def test_open_ro__uncompressed():
    handle = open_ro(test_file('fasta_file.fasta'))
    try:
        assert_equal(handle.read(),
                     b'>This_is_FASTA!\nACGTN\n'
                     b'>This_is_ALSO_FASTA!\nCGTNA\n')
    finally:
        handle.close()


def test_open_ro__gz():
    handle = open_ro(test_file('fasta_file.fasta.gz'))
    try:
        assert_equal(handle.read(),
                     b'>This_is_GZipped_FASTA!\nACGTN\n'
                     b'>This_is_ALSO_GZipped_FASTA!\nCGTNA\n')
    finally:
        handle.close()


def test_open_ro__bz2():
    handle = open_ro(test_file('fasta_file.fasta.bz2'))
    try:
        assert_equal(handle.read(),
                     b'>This_is_BZ_FASTA!\nCGTNA\n'
                     b'>This_is_ALSO_BZ_FASTA!\nACGTN\n')
    finally:
        handle.close()


class OddException(RuntimeError):
    pass


def test_open_ro__close_handle_on_error():
    class _FileMock(object):
        def __init__(self, filename):
            self._close_called = False
            assert_equal(filename, "/var/abc")

        def read(self, *_args, **_kwargs):
            # pylint: disable=R0201
            raise OddException("ARGH!")

        def close(self):
            self._close_called = True

    try:
        paleomix.common.fileutils.open = _FileMock
        assert_raises(OddException, open_ro, "/var/abc")
    finally:
        del paleomix.common.fileutils.open


###############################################################################
###############################################################################
# Tests for 'try_remove'

@with_temp_folder
def test_try_remove(temp_folder):
    fpath = os.path.join(temp_folder, "test.txt")
    set_file_contents(fpath, "1 2 3")
    assert try_remove(fpath)
    assert not os.path.exists(fpath)


@with_temp_folder
def test_try_remove__missing(temp_folder):
    fpath = os.path.join(temp_folder, "test.txt")
    assert not try_remove(fpath)
    assert not os.path.exists(fpath)


@with_temp_folder
def test_try_remove__non_file(temp_folder):
    assert_raises(OSError, try_remove, temp_folder)


###############################################################################
###############################################################################
# Tests for 'try_rmdir'

@with_temp_folder
def test_try_rmdir(temp_folder):
    fpath = os.path.join(temp_folder, "testdir")
    os.mkdir(fpath)
    assert try_rmdir(fpath)
    assert not os.path.exists(fpath)


@with_temp_folder
def test_try_rmdir__missing(temp_folder):
    fpath = os.path.join(temp_folder, "testdir")
    assert not try_rmdir(fpath)
    assert not os.path.exists(fpath)


@with_temp_folder
def test_try_rmdir__non_file(temp_folder):
    fpath = os.path.join(temp_folder, "test.txt")
    set_file_contents(fpath, "1 2 3")
    assert_raises(OSError, try_rmdir, fpath)


###############################################################################
###############################################################################
# Tests for 'try_rmdir'

@with_temp_folder
def test_try_rmtree(temp_folder):
    fpath = os.path.join(temp_folder, "testdir")
    os.mkdir(fpath)
    set_file_contents(os.path.join(fpath, "file"), "1 2 3")
    assert try_rmtree(fpath)
    assert not os.path.exists(fpath)


@with_temp_folder
def test_try_treedir__missing(temp_folder):
    fpath = os.path.join(temp_folder, "testdir")
    assert not try_rmtree(fpath)
    assert not os.path.exists(fpath)


###############################################################################
###############################################################################
# Tests for 'describe_files'

def test_describe_files__no_files():
    assert_equal(describe_files(()), "No files")


def test_describe_files__single_file():
    fpath = "/var/foo/bar"
    assert_equal(describe_files((fpath,)), repr(fpath))


def test_describe_files__same_path_abs__3_differences():
    fpaths = ("/var/foo/bar", "/var/foo/foo")
    assert_equal(describe_files(fpaths), "2 files in '/var/foo'")


def test_describe_files__same_path_abs__2_differences():
    fpaths = ("/var/foo/faz", "/var/foo/foo")
    assert_equal(describe_files(fpaths), "'/var/foo/f??'")


def test_describe_files__same_path_abs__1_differences():
    fpaths = ("/var/foo/faz", "/var/foo/fao")
    assert_equal(describe_files(fpaths), "'/var/foo/fa?'")


def test_describe_files__different_paths_abs():
    fpaths = ("/var/foo/bar", "/var/bar/foo")
    assert_equal(describe_files(fpaths), "2 files")


def test_describe_files__same_path_rel():
    fpaths = ("var/foo/bar", "var/foo/foo")
    assert_equal(describe_files(fpaths), "2 files in 'var/foo'")


def test_describe_files__different_paths_rel():
    fpaths = ("var/foo/bar", "var/bar/foo")
    assert_equal(describe_files(fpaths), "2 files")


def test_describe_files__iterable():
    fpaths = iter(("/var/foo/bar", "/var/foo/foo"))
    assert_equal(describe_files(fpaths), "2 files in '/var/foo'")


def test_describe_files__none_files():
    assert_raises(ValueError, describe_files, None)


###############################################################################
###############################################################################
# Tests for 'describe_paired_files'

def test_describe_paired_files__single_file():
    fpath = "/var/foo/bar"
    assert_equal(describe_paired_files((fpath,), ()), repr(fpath))


def test_describe_paired_files__identical_files():
    fpath = "/var/foo/bar"
    ftuple = (fpath,)
    assert_equal(describe_paired_files(ftuple, ftuple), repr(fpath))


def test_describe_paired_files__same_path__similar_files():
    files_1 = ("foo/1_abc", "foo/1_def")
    files_2 = ("foo/1_ghi", "foo/1_jkl")
    expected = "'foo/1_???'"
    result = describe_paired_files(files_1, files_2)
    assert_equal(result, expected)


def test_describe_paired_files__same_path__similar_files__different_prefixes():
    files_1 = ("foo/1_abc", "foo/1_def")
    files_2 = ("foo/2_ghi", "foo/2_jkl")
    expected = "'foo/[12]_???'"
    result = describe_paired_files(files_1, files_2)
    assert_equal(result, expected)


def test_describe_paired_files__same_path__similar_files__too_different():
    files_1 = ("foo/1a_abc", "foo/1a_def")
    files_2 = ("foo/2b_ghi", "foo/2b_jkl")
    expected = "2 pair(s) of files in 'foo'"
    result = describe_paired_files(files_1, files_2)
    assert_equal(result, expected)


def test_describe_paired_files__same_path__different_files():
    files_1 = ("foo/1_abc", "foo/2_def")
    files_2 = ("foo/3_ghi", "foo/4_jkl")
    expected = "2 pair(s) of files in 'foo'"
    result = describe_paired_files(files_1, files_2)
    assert_equal(result, expected)


def test_describe_paired_files__same_path__different_file_lens():
    files_1 = ("foo/1_a", "foo/2_de")
    files_2 = ("foo/3_g", "foo/4_jk")
    expected = "2 pair(s) of files in 'foo'"
    result = describe_paired_files(files_1, files_2)
    assert_equal(result, expected)


def test_describe_paired_files__different_path_and_files():
    files_1 = ("foo/1_abc", "bar/2_def")
    files_2 = ("zed/3_ghi", "not/4_jkl")
    expected = "2 pair(s) of files"
    result = describe_paired_files(files_1, files_2)
    assert_equal(result, expected)


def test_describe_paired_files__files_1_longer():
    assert_raises(ValueError, describe_paired_files, ("a", "b"), ("c",))


def test_describe_paired_files__files_2_longer():
    assert_raises(ValueError, describe_paired_files, ("a",), ("b", "c"))


def test_describe_paired_files__none_files():
    assert_raises(ValueError, describe_paired_files, None, None)


def test_describe_paired_files__none_files_1():
    assert_raises(ValueError, describe_paired_files, None, ())


def test_describe_paired_files__none_files_2():
    assert_raises(ValueError, describe_paired_files, (), None)



#!/usr/bin/env python
#
# Copyright (c) 2013 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import codecs
import os
import sys

from setuptools import setup, find_packages


if (sys.version_info.major != 2) or (sys.version_info.minor != 7):
    sys.stderr.write("ERROR: Python version 2.7.x required!\n")
    sys.stderr.write("       Current version is v%s\n"
                     % (sys.version.replace("\n", " "),))
    sys.exit(1)


def _get_version():
    """Retrieve version from current install directory."""
    env = {}
    with open(os.path.join("paleomix", "__init__.py")) as handle:
        exec(handle.read(), env)

    return env["__version__"]


def _get_readme():
    """Retrieves contents of README.rst, forcing UTF-8 encoding."""
    with codecs.open("README.rst", encoding="utf-8") as handle:
        return handle.read()


setup(
    name='paleomix',
    version=_get_version(),

    description='Bioinformatics pipelines for HTS data',
    long_description=_get_readme(),

    url='https://github.com/MikkelSchubert/paleomix',

    author='Mikkel Schubert',
    author_email='MSchubert@snm.ku.dk',

    license='MIT',

    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2 :: Only',
        'Programming Language :: Python :: 2.7',
    ],

    keywords='pipeline bioinformatics hts phylogeny bam',

    packages=find_packages(exclude=['misc', 'tests']),

    install_requires=['pysam>=0.8.3'],

    # Dependencies set in setup_requires to allow use of 'setup.py nosetests'
    setup_requires=['nose>=1.3.0',
                    'flexmock>=0.9.7',
                    'coverage>=4.0.0'],

    test_suite='nose.collector',

    entry_points={
        'console_scripts': [
            'paleomix=paleomix:run',

            # Aliases used in previous publications
            'bam_pipeline=paleomix:run_bam_pipeline',
            'bam_rmdup_collapsed=paleomix:run_rmdup_collapsed',
            'conv_gtf_to_bed=paleomix:run_gtf_to_bed',
            'phylo_pipeline=paleomix:run_phylo_pipeline',
            'trim_pipeline=paleomix:run_trim_pipeline',
        ],
    },

    zip_safe=False,

    include_package_data=True,
)
